"""Test for RSMDef.py - Rural Site & Vertical Diffusion Model (RSM & VDM)"""

import os
import pytest
from .test_base import auto_setup_uwg, calculate_tolerance, setup_open_matlab_ref
from functools import reduce


TEST_DIR = os.path.abspath(os.path.dirname(__file__))


def test_rsm_init():
    """Test Initialize RSM instance for rural site parameters

    Reference init:

    lat = 1.37
    lon = 103.98
    GMT = 8.0
    rural_height = 0.1            # average obstacle height from initialize.uwg
    urban_height = 1.0            # average building height / 10.
    T_init = 297.85               # initial dry bulb
    P_init = 100900.0             # initial pressure

    testuwg.geoParam # geographic parameters

    RSM = RSMDef(
        lat, lon, GMT, rural_height, T_init, P_init, geo_param, self.RESOURCE_PATH)
    """

    testuwg = auto_setup_uwg()
    testuwg.generate()
    testuwg.RSM.__repr__()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay / 3600. == pytest.approx(0.0, abs=1e-15)

    # Test z_meso list lenght
    assert len(testuwg.RSM.z_meso) == pytest.approx(56., abs=1e-6)
    assert testuwg.RSM.dz[1] == pytest.approx(4.4, abs=1e-6)
    assert testuwg.RSM.z[1] == pytest.approx(6.2, abs=1e-6)

    # Test self.nz0, self.nzref, self.nzfor, self.nz10, self.nzi
    assert testuwg.RSM.nz0 == pytest.approx(1, abs=1e-6)
    assert testuwg.RSM.nzref == pytest.approx(17, abs=1e-6)
    assert testuwg.RSM.nzfor == pytest.approx(13., abs=1e-6)
    assert testuwg.RSM.nz10 == pytest.approx(3, abs=1e-6)
    assert testuwg.RSM.nzi == pytest.approx(35, abs=1e-6)

    # Test the tempProfile
    assert len(testuwg.RSM.tempProf) == pytest.approx(17, abs=1e-6)
    assert testuwg.RSM.tempProf[16] == pytest.approx(297.8500, abs=1e-6)

    # Test the presProfile with values from UWG_Matlab to 15 digits of precision
    matlab_presProf = [
        1.009000000000000, 1.008490604570617, 1.007930481749694, 1.007314603288614,
        1.006637447436641, 1.005892951541777, 1.005074460319214, 1.004174669438980,
        1.003185564067109, 1.002098351979773, 1.000903390854001, 0.999590109338784,
        0.998146921496037, 0.996561134212841, 0.994818847169896, 0.992904845102566,
        0.990802481868362]  # * 1.01e+05

    # Test varying precision manually
    assert testuwg.RSM.presProf[0] == pytest.approx(1.009e5, abs=1e-15)
    assert testuwg.RSM.presProf[-1] == pytest.approx(0.990802481868362e5, abs=1e-10)
    # Test all
    for i in range(1, len(matlab_presProf) - 1):
        tol = calculate_tolerance(matlab_presProf[i] * 1e5, 15.0)
        assert testuwg.RSM.presProf[i] == pytest.approx(matlab_presProf[i]*1e5, abs=tol)

    # Test the tempRealProf
    # Add dynamic tolerance
    assert testuwg.RSM.tempRealProf[0] == pytest.approx(2.978499999999999*1e2, abs=1e-10)
    assert testuwg.RSM.tempRealProf[6] == pytest.approx(2.975182902489641*1e2, abs=1e-10)
    assert testuwg.RSM.tempRealProf[-1] == \
        pytest.approx(2.963044480678944*1e2, abs=1e-10)

    # Test the densityProfS
    matlab_densityProfS = [
        1.180352339267655, 1.180149676936383, 1.179703870430437, 1.179213600207828,
        1.178674444441198, 1.178081544270841, 1.177429561181760, 1.176712630344852,
        1.175924309566055, 1.175057523461647, 1.174104502451500, 1.173056716134468,
        1.171904800590439, 1.170638479118457, 1.169246475901031, 1.167716422060640,
        1.166034753626033, 1.165110236615915]

    for i in range(len(matlab_densityProfS)):
        tol = calculate_tolerance(matlab_densityProfS[i] * 1e5, 15.0)
        assert testuwg.RSM.densityProfS[i] == \
            pytest.approx(matlab_densityProfS[i], abs=tol)

    assert len(testuwg.RSM.densityProfC) == pytest.approx(testuwg.RSM.nzref, abs=1e-15)
    assert len(testuwg.RSM.densityProfS) == pytest.approx(testuwg.RSM.nzref+1, abs=1e-15)


def test_rsm_vdm():
    """ test RSM VDM against matlab references."""

    testuwg = auto_setup_uwg()

    # Test Jan 1 (winter, no vegetation coverage)
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 1

    testuwg.generate()

    # In order to avoid integration effects. Test only first time step
    # Subtract timestep to stop at 300 sec
    testuwg.simTime.nt -= (23 * 12 + 11)

    # Run simulation
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    # Matlab Checking for RSM.VDM
    # 2d matrix = 7 x 16
    uwg_python_val = [
        testuwg.RSM.presProf,
        testuwg.RSM.tempRealProf,
        testuwg.RSM.densityProfC,
        testuwg.RSM.densityProfS,
        testuwg.RSM.tempProf,
        testuwg.RSM.windProf,
        [testuwg.RSM.ublPres]]

    # Flatten 2d matrix into 1d vector
    uwg_python_val = reduce(lambda x, y: x + y, uwg_python_val)

    # Matlab checking
    uwg_matlab_val = setup_open_matlab_ref("matlab_rsmdef", "matlab_ref_rsmdef_vdm.txt")

    # Matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        # print uwg_python_val[i], uwg_matlab_val[i]
        tol = calculate_tolerance(uwg_python_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            "error at index={}".format(i)


def test_rsm_dissipation_bougeault():
    """Test RSM VDM against matlab references."""

    epw_path = os.path.join(TEST_DIR, 'epw', 'CAN_ON_Toronto.716240_CWEC.epw')
    param_path = os.path.join(TEST_DIR, 'parameters', 'initialize_toronto.uwg')
    testuwg = auto_setup_uwg(epw_path=epw_path, param_path=param_path)
    testuwg.month = 6
    testuwg.day = 1
    testuwg.nday = 30

    # run simulation
    testuwg.generate()
    testuwg.simulate()

    # check date
    # print testuwg.simTime
    assert testuwg.simTime.month == 7
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(0.0, abs=1e-15)

    # dld, dlu values after 1 day
    uwg_python_val = [
        testuwg.RSM.dlu,
        testuwg.RSM.dld]

    # Flatten 2d matrix into 1d vector
    uwg_python_val = reduce(lambda x, y: x + y, uwg_python_val)

    # Matlab checking
    uwg_matlab_val = setup_open_matlab_ref(
        "matlab_rsmdef", "matlab_rsmdef_dissipation_bougeault.txt")

    # Matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_python_val[i], 11.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            "error at index={}".format(i)



"""Class for Rural Site Model (RSM) and Vertical Diffusion Model (VDM)."""
from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import os
from math import sqrt, pow, log
from .utilities import is_near_zero

# TODO: Method and arg docstrings are incomplete. Should be completed by someone who
# understands RSM and VSM dynamics.


class RSMDef(object):
    """Rural Site Model (RSM) and Vertical Diffusion Model (VDM).

    This class calculates the vertical profiles of air temperature above the weather
    station[1].

    Note:
        [1] 'The uwg' (2012) Eq. 4, 5, 6.

    Args:
        lat: Number for latitude in degrees.
        lon: Number for longitude in degrees.
        gmt: Number for GMT hour correction.
        height: Number for rural average obstacle height in meters.
        T_init: Number for initial dry bulb temperature.
        P_init: Number for initial pressure.
        parameter: Param object for geographic parameters.
        z_meso_path: Text string for height reference data filepath for mesoscale
            component.

    Properties
        * z_meso -- list of mesoscale heights.
        * lat -- latitude [deg]
        * lon -- longitude [deg]
        * gmt -- GMT hour correction
        * height -- average obstacle height [m]
        * z0r -- rural roughness length [m]
        * disp -- rural displacement lenght [m]
        * z -- vertical height [m]
        * dz -- vertical discretization [m]
        * nz0 -- layer number at zmt [m]
        * nzref -- layer number at zref [m]
        * nzfor -- layer number at zfor [m]
        * nz10 -- layer number at zmu [m]
        * nzi -- layer number at zi_d [m]
        * tempProf -- potential temperature profile at the rural site [K]
        * presProf -- pressure profile at the rural site [Pa]
        * tempRealProf -- real temperature profile at the rural site [K]
        * densityProfC -- density profile at the center of layers [kg m-3]
        * densityProfS -- density profile at the sides of layers [kg m-3]
        * windProf -- wind profile at the rural site [m s-1]
        * ublPres -- average pressure at UBL [Pa]
    """

    def __init__(self, lat, lon, gmt, height, T_init, P_init, parameter, z_meso_path):

        self.z_meso = RSMDef.load_z_meso(z_meso_path)
        self.lat = lat
        self.lon = lon
        self.gmt = gmt
        self.height = height
        self.z0r = 0.1 * height
        self.disp = 0.5 * height

        # vertical grid at the rural site
        self.z = [0 for x in range(len(self.z_meso)-1)]
        self.dz = [0 for x in range(len(self.z_meso)-1)]

        # Initialize empty parameters to compute for RSMDef
        self.nz0 = None
        self.nzref = None
        self.nzfor = None
        self.nz10 = None
        self.nzi = None
        self.tempProf = None
        self.presProf = None
        self.tempRealProf = None
        self.densityProfC = None
        self.densityProfS = None
        self.windProf = None
        self.ublPres = None

        for zi in range(len(self.z_meso)-1):
            self.z[zi] = 0.5 * (self.z_meso[zi] + self.z_meso[zi+1])
            self.dz[zi] = self.z_meso[zi+1] - self.z_meso[zi]

        # Define initial booleans
        ll = True
        mm = True
        nn = True
        oo = True
        pp = True

        # Define self.nz0, self.nzref, self.nzfor, self.nz10, self.nzi
        for iz in range(len(self.z_meso)-1):
            # self.nz0: self.z index >= reference height for weather station
            eq_th = is_near_zero(self.z[iz] - parameter.tempHeight)
            if (eq_th or self.z[iz] > parameter.tempHeight) and ll:
                self.nz0 = iz + 1  # layer number at zmt (m)
                ll = False

            # self.nzref: self.z index >= reference inversion height
            eq_rh = is_near_zero(self.z[iz] - parameter.refHeight)
            if (eq_rh or self.z[iz] > parameter.refHeight) and mm:
                self.nzref = iz + 1  # layer number at zref (m)
                mm = False

            # self.nzfor: self.z index >= nighttime boundary layer height
            eq_nh = is_near_zero(self.z[iz] - parameter.nightBLHeight)
            if (eq_nh or self.z[iz] > parameter.nightBLHeight) and nn:
                self.nzfor = iz + 1   # layer number at zfor (m)
                nn = False

            # self.nz10: self.z index >= wind height
            eq_wh = is_near_zero(self.z[iz] - parameter.windHeight)
            if (eq_wh or self.z[iz] > parameter.windHeight) and oo:
                self.nz10 = iz + 1  # layer number at zmu (m)
                oo = False

            eq_dh = is_near_zero(self.z[iz] - parameter.dayBLHeight)
            if (eq_dh or self.z[iz] > parameter.dayBLHeight) and pp:
                self.nzi = iz + 1  # layer number at zi_d (m)
                pp = False

        # Define temperature, pressure and density vertical profiles
        self.tempProf = [T_init for x in range(self.nzref)]
        self.presProf = [P_init for x in range(self.nzref)]

        for iz in range(1, self.nzref):
            self.presProf[iz] = (
                (self.presProf[iz-1] ** (parameter.r / parameter.cp) - parameter.g /
                 parameter.cp * (P_init ** (parameter.r / parameter.cp)) *
                 (1. / self.tempProf[iz] + 1. / self.tempProf[iz-1]) * 0.5 *
                 self.dz[iz]) ** (1. / (parameter.r / parameter.cp)))

        self.tempRealProf = [T_init for x in range(self.nzref)]
        for iz in range(self.nzref):
            self.tempRealProf[iz] = (
                self.tempProf[iz] * (self.presProf[iz] / P_init) **
                (parameter.r / parameter.cp))

        self.densityProfC = [None for x in range(self.nzref)]
        for iz in range(self.nzref):
            self.densityProfC[iz] = \
                self.presProf[iz] / parameter.r / self.tempRealProf[iz]

        self.densityProfS = [self.densityProfC[0]
                             for x in range(self.nzref + 1)]

        for iz in range(1, self.nzref):
            self.densityProfS[iz] = (
                (self.densityProfC[iz] * self.dz[iz-1] + self.densityProfC[iz-1] *
                 self.dz[iz]) / (self.dz[iz-1]+self.dz[iz]))

        self.densityProfS[self.nzref] = self.densityProfC[self.nzref-1]
        self.windProf = [1 for x in range(self.nzref)]

    def vdm(self, forc, rural, parameter, simTime):
        """Vertical Diffusion Model[1].

        Note:
        [1] 'The uwg' (2012), Eq. (4)
        """

        self.tempProf[0] = forc.temp    # Lower boundary condition

        # compute pressure profile
        for iz in reversed(list(range(self.nzref))[1:]):
            self.presProf[iz-1] = (
                (pow(self.presProf[iz], parameter.r / parameter.cp) +
                 parameter.g / parameter.cp *
                 (pow(forc.pres, parameter.r / parameter.cp)) *
                 (1./self.tempProf[iz] + 1./self.tempProf[iz-1]) *
                 0.5 * self.dz[iz]) ** (1. / (parameter.r / parameter.cp)))

        # compute the real temperature profile
        for iz in range(self.nzref):
            self.tempRealProf[iz] = (
                self.tempProf[iz] * (self.presProf[iz] / forc.pres) **
                (parameter.r / parameter.cp))

        # compute the density profile
        for iz in range(self.nzref):
            self.densityProfC[iz] = \
                self.presProf[iz] / parameter.r / self.tempRealProf[iz]
        self.densityProfS[0] = self.densityProfC[0]

        for iz in range(1, self.nzref):
            self.densityProfS[iz] = (
                (self.densityProfC[iz] * self.dz[iz-1] + self.densityProfC[iz-1] *
                 self.dz[iz]) / (self.dz[iz-1] + self.dz[iz]))

        self.densityProfS[self.nzref] = self.densityProfC[self.nzref-1]

        # Ref: The uwg (2012), Eq. (5)
        # compute diffusion coefficient
        cd, ustarRur = self.diffusion_coefficient(
            self.densityProfC[0], self.z, self.dz, self.z0r, self.disp, self.tempProf[0],
            rural.sens, self.nzref, forc.wind, self.tempProf, parameter)

        # solve diffusion equation
        self.tempProf = RSMDef.diffusion_equation(
            self.nzref, simTime.dt, self.tempProf, self.densityProfC, self.densityProfS,
            cd, self.dz)

        # compute wind profile
        # N.B In Matlab, negative values are converted to complex values.
        # log(-x) = log(x) + log(-1) = log(x) + i*pi
        # Python will throw an exception. Negative value occurs here if
        # VDM is run for average obstacle height ~ 4m.
        for iz in range(self.nzref):
            self.windProf[iz] = \
                ustarRur / parameter.vk * \
                log((self.z[iz] - self.disp) / self.z0r)

        # Average pressure
        self.ublPres = 0.
        for iz in range(self.nzfor):
            self.ublPres = (
                self.ublPres + self.presProf[iz] * self.dz[iz] /
                (self.z[self.nzref-1] + self.dz[self.nzref-1] / 2.))

    def diffusion_coefficient(self, rho, z, dz, z0, disp, tempRur, heatRur, nz, uref, th,
                              parameter):
        # Initialization
        Kt = [0 for x in range(nz+1)]
        ws = [0 for x in range(nz)]
        te = [0 for x in range(nz)]
        # Friction velocity (Louis 1979)
        ustar = parameter.vk * uref/log((10.-disp)/z0)

        # Monin-Obukhov length
        lengthRur = max(-rho * parameter.cp * ustar ** 3 * tempRur / parameter.vk /
                        parameter.g / heatRur, -50.)

        if heatRur > 1e-2:
            # Unstable conditions

            # Convective velocity scale
            wstar = (
                (parameter.g * heatRur * parameter.dayBLHeight / rho / parameter.cp /
                    tempRur) ** (1 / 3.))
            # Wind profile function
            phi_m = (1-8. * 0.1 * parameter.dayBLHeight /
                     lengthRur) ** (-1. / 3.)

            for iz in range(nz):
                # Mixed-layer velocity scale
                ws[iz] = (
                    (ustar ** 3 + phi_m * parameter.vk * wstar ** 3 * z[iz] /
                        parameter.dayBLHeight) ** (1 / 3.))

                # TKE approximation
                te[iz] = max(ws[iz] ** 2., 0.01)

        else:
            # Stable and neutral conditions
            for iz in range(nz):
                # TKE approximation
                te[iz] = max(ustar ** 2., 0.01)

        # lenght scales (l_up, l_down, l_k, l_eps)
        self.dlu, self.dld = RSMDef.dissipation_bougeault(
            parameter.g, nz, z, dz, te, th)
        self.dld, dls, dlk = RSMDef.length_bougeault(nz, self.dld, self.dlu, z)

        # Boundary-layer diffusion coefficient
        for iz in range(nz):
            Kt[iz] = 0.4 * dlk[iz] * sqrt(te[iz])

        Kt[nz] = Kt[nz-1]

        return Kt, ustar

    @staticmethod
    def diffusion_equation(nz, dt, co, da, daz, cd, dz):

        cddz = [0 for i in range(nz + 2)]
        a = [[0 for j in range(3)] for i in range(nz)]
        c = [0 for i in range(nz)]

        cddz[0] = daz[0] * cd[0] / dz[0]
        for iz in range(1, nz):
            cddz[iz] = 2. * daz[iz] * cd[iz] / (dz[iz] + dz[iz-1])
        cddz[nz] = daz[nz] * cd[nz] / dz[nz]

        a[0][0] = 0.
        a[0][1] = 1.
        a[0][2] = 0.
        c[0] = co[0]

        for iz in range(1, nz - 1):
            dzv = dz[iz]
            a[iz][0] = -cddz[iz] * dt / dzv / da[iz]
            a[iz][1] = 1 + dt * (cddz[iz] + cddz[iz+1]) / dzv / da[iz]
            a[iz][2] = -cddz[iz+1] * dt / dzv / da[iz]
            c[iz] = co[iz]

        a[nz-1][0] = -1.
        a[nz-1][1] = 1.
        a[nz-1][2] = 0.
        c[nz-1] = 0.

        return RSMDef.invert(nz, a, c)

    @staticmethod
    def dissipation_bougeault(g, nz, z, dz, te, pt):
        # N.B on translation from UWG_Matlab
        # list length (i.e nz) != list indexing (i.e dlu[0] in python
        # wherease in matlab it is

        dlu = [0 for x in range(nz)]
        dld = [0 for x in range(nz)]

        for iz in range(nz):
            zup = 0.0
            dlu[iz] = z[nz] - z[iz] - (dz[iz] / 2.)
            zzz = 0.0
            zup_inf = 0.0
            beta = g / pt[iz]

            for izz in range(iz, nz - 1):
                dzt = (dz[izz+1] + dz[izz]) / 2.
                zup = zup - beta * pt[iz] * dzt
                zup = zup + beta * (pt[izz+1] + pt[izz]) * dzt / 2.
                zzz = zzz + dzt

                _chkzero = is_near_zero(te[iz] - zup_inf)
                if (te[iz] < zup) and ((te[iz] > zup_inf) or _chkzero):
                    bbb = (pt[izz+1] - pt[izz]) / dzt

                    if not is_near_zero(bbb - 0.):
                        _t1 = (
                            sqrt(max(0., (beta * (pt[izz] - pt[iz])) ** 2. +
                                     2. * bbb * beta * (te[iz] - zup_inf))))
                        tl = (-beta * (pt[izz] - pt[iz]) + _t1) / bbb / beta
                    else:
                        tl = (te[iz] - zup_inf) / (beta * (pt[izz] - pt[iz]))
                    dlu[iz] = max(1., zzz - dzt + tl)
                zup_inf = zup

            zdo = 0.
            zdo_sup = 0.
            dld[iz] = z[iz] + dz[iz] / 2.
            zzz = 0.

            for izz in range(iz, 0, -1):
                dzt = (dz[izz-1] + dz[izz]) / 2.
                zdo = zdo + beta * pt[iz] * dzt
                zdo = zdo - beta * (pt[izz - 1] + pt[izz]) * dzt / 2.
                zzz = zzz + dzt

                _chkzero = is_near_zero(te[iz] - zdo_sup)
                if (te[iz] < zdo) and ((te[iz] > zdo_sup) or _chkzero):

                    bbb = (pt[izz] - pt[izz-1]) / dzt

                    if not is_near_zero(bbb - 0.):
                        tl = (
                            (beta * (pt[izz] - pt[iz]) +
                             sqrt(max(0., (beta * (pt[izz] - pt[iz])) ** 2. +
                                      2. * bbb * beta * (te[iz] - zdo_sup)))) / bbb / beta)
                    else:
                        tl = (te[iz] - zdo_sup) / (beta * (pt[izz] - pt[iz]))
                    dld[iz] = max(1., zzz - dzt + tl)
                zdo_sup = zdo

        return dlu, dld

    @staticmethod
    def length_bougeault(nz, dld, dlu, z):

        dlg = [0 for x in range(nz)]
        dls = [0 for x in range(nz)]
        dlk = [0 for x in range(nz)]

        for iz in range(nz):
            dlg[iz] = (z[iz] + z[iz+1]) / 2.

        for iz in range(nz):
            dld[iz] = min(dld[iz], dlg[iz])
            dls[iz] = sqrt(dlu[iz] * dld[iz])
            dlk[iz] = min(dlu[iz], dld[iz])

        return dld, dls, dlk

    @staticmethod
    def invert(nz, A, C):
        """Inversion and resolution of a tridiagonal matrix A X = C.

        nz number of layers
        a(*,1) lower diagonal (Ai, i-1)
        a(*,2) principal diagonal (Ai, i)
        a(*,3) upper diagonal (Ai, i+1)
        """

        X = [0 for i in range(nz)]

        for i in reversed(range(nz - 1)):
            C[i] = C[i] - A[i][2] * C[i+1] / A[i+1][1]
            A[i][1] = A[i][1] - A[i][2] * A[i+1][0] / A[i+1][1]

        for i in range(1, nz, 1):
            C[i] = C[i] - A[i][0] * C[i-1] / A[i-1][1]

        for i in range(nz):
            X[i] = C[i] / A[i][1]

        return X

    @staticmethod
    def load_z_meso(z_meso_path):
        """Get list of mesoscale height reference data."""

        assert os.path.exists(z_meso_path), \
            "File: '{}' does not exist.".format(z_meso_path)

        z_meso = []
        with open(z_meso_path, 'r') as f:
            for txtline in f:
                # Strip white spaces, change to float
                z = float(''.join(txtline.split()))
                z_meso.append(z)

        return z_meso

    def __repr__(self):
        return 'RSM,\n lat: {}\n lon: {}\n gmt: {}\n height: {}\n z0r: {}\n ' \
            'disp: {}'.format(self.lat, self.lon, self.gmt, self.height, self.z0r,
                              self.disp)



import os
import re

fname = "batch_input.txt"
curr_dir = os.path.abspath(os.path.dirname(__file__))
fpath = os.path.join(curr_dir,fname)

def batch_file_process():
    f = open(fpath,'r')
    L = f.readlines()
    f.close()

    # Clean raw data
    prefix2delete = ""
    prefix2add = ""
    objectref = ""
    methodref = ""
    i = 0
    while i < len(L):
        line = L[i].split(" ")[0]
        line = "".join(line.split()) # removes all white space from string

        if "prefix2delete" in line:
            prefix2delete = line.split("=")[-1]
            L.pop(i)
        elif "prefix2add" in line:
            prefix2add = line.split("=")[-1]
            L.pop(i)
        elif "objectref" in line:
            objectref = line.split("=")[-1]
            L.pop(i)
        elif "methodref" in line:
            methodref = line.split("=")[-1]
            L.pop(i)
        elif line == "":
            L.pop(i)
        else:
            line = line.split(",")[0] if "," in line else line

            L[i] = line
            i += 1

    # For uwg_python_val
    #for i in xrange(len(L)):
    #    L[i] = L[i].split(prefix2delete)[-1] if prefix2delete!="" else L[i]
    #    print '{prefix}.{valuename},'.format(valuename=L[i], prefix=prefix2add)

    # create three outputs
    print "%%{"
    print "fileID = fopen('..\\UWG_Python\\tests\\matlab_ref\\matlab_{}\\matlab_ref_{}.txt','w');".format(objectref,objectref+"_"+methodref)
    print "format long;"
    for i in xrange(len(L)):
        L[i] = L[i].split(prefix2delete)[-1] if prefix2delete!="" else L[i]
        print 'fprintf(fileID, "{value}", {prefix}.{valuename});'.format(value=r"%.16f\n", valuename=L[i], prefix=prefix2add)
    print "fclose(fileID);"
    print "%}%"


if __name__ == "__main__":
    batch_file_process()

"""
ARCHIVE_RSM_VDM:
self.uwg.rural.aeroCond        # Convection coef (refL uwg, eq.12)
self.uwg.rural.waterStorage    # thickness of water film (m) (only for horizontal surfaces)
self.uwg.rural.solAbs          # solar radiation absorbed (W m-2)
self.uwg.rural.lat             # surface latent heat flux (W m-2)
self.uwg.rural.sens            # surface sensible heat flux (W m-2)
self.uwg.rural.flux             # external surface heat flux (W m-2)"

prefix2delete=self.uwg.rural.
prefix2add=RSM
objectref=rsmdef
methodref=surfflux
"""



"""Tests for UBLDef object."""

import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref, calculate_tolerance
from functools import reduce


def test_ubl_init():
    """Test ubl constructor."""

    testuwg = auto_setup_uwg()
    testuwg.generate()
    testuwg.UBL.__repr__()

    # Get uwg_python values
    uwg_python_val = [
        # characteristic length of the urban area (m)
        testuwg.UBL.charLength,
        testuwg.UBL.perimeter,
        # horizontal urban area (m2)
        testuwg.UBL.urbArea,
        # length of the side of the urban area orthogonal to wind dir (m)
        testuwg.UBL.orthLength,
        # length of the side of the urban area parallel to wind dir (m)
        testuwg.UBL.paralLength,
        # urban boundary layer temperature (K)
        testuwg.UBL.ublTemp,
        # urban boundary layer temperature discretization (K)
        reduce(lambda x, y: x + y, testuwg.UBL.ublTempdx),
        testuwg.UBL.dayBLHeight,  # daytime mixing height, orig = 700
        # Sing: 80, Bub-Cap: 50, nighttime boundary-layer height (m); orig 80
        testuwg.UBL.nightBLHeight]

    uwg_matlab_val = testuwg = setup_open_matlab_ref(
        'matlab_ubl', 'matlab_ref_ubl_init.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val), \
        'matlab={}, python={}'.format(len(uwg_matlab_val), len(uwg_python_val))

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_python_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], tol), \
            'error at index={}'.format(i)


def test_ublmodel():
    """Test ubl constructor."""

    testuwg = auto_setup_uwg()

    # Test Jan 1 (winter, no vegetation coverage)
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 1

    # set_input
    testuwg.generate()

    # In order to avoid integration effects. Test only first time step
    # Subtract timestep to stop at 300 sec
    testuwg.simTime.nt -= (23 * 12 + 11)

    # Run simulation
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    # Get uwg_python values
    uwg_python_val = [
        # urban boundary layer temperature (K)
        testuwg.UBL.ublTemp,
        # urban boundary layer temperature discretaization (K)
        reduce(lambda x, y: x + y, testuwg.UBL.ublTempdx),
        # night boundary layer height (m)
        testuwg.UBL.dayBLHeight,
        # night boundary layer height (m)
        testuwg.UBL.nightBLHeight]

    uwg_matlab_val = setup_open_matlab_ref(
        'matlab_ubl', 'matlab_ref_ublmodel.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_python_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], tol), \
            'error at index={}'.format(i)



"""Test tree and vegetation coverage."""
from __future__ import division

import os
import pytest
from uwg import UWG, solarcalcs


EPW_PATH = \
    os.path.join(os.path.dirname(__file__), 'epw',
                 'SGP_Singapore.486980_IWEC.epw')


def test_tree_vegetation_ratio():
    """Test road vegetation fractions."""

    model = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', grasscover=0,
        treecover=0, epw_path=EPW_PATH)

    with pytest.raises(AssertionError):
        model.blddensity = 0.5
        model.grasscover = 0.5
        model.treecover = 0.1  # > grasscover

    with pytest.raises(AssertionError):
        model.blddensity = 0.5
        model.treecover = 0.3
        model.grasscover = 0.3  # > treecover

    with pytest.raises(AssertionError):
        model.treecover = 0.51
        model.grasscover = 0.49
        model.blddensity = 0.001  # > veg + grass

    model.grasscover = 0.125
    model.treecover = 0.125

    # generate road
    model.generate()

    assert model.vegcover == pytest.approx(0.25, 1e-10)
    # test road coverage (vegcover / (1 - density))
    assert model.blddensity == pytest.approx(0.5, 1e-10)
    # (0.125 + 0.125)/0.5 = 0.5
    assert model.road.vegcoverage == pytest.approx(0.5, 1e-10)
    # 0.125/0.5 = 1/8 * 2 = 1/4
    assert model.UCM.roadShad == pytest.approx(0.25, 1e-10)


def test_tree_drybulb():
    """Test if drybulb decreases when tree cover increases."""

    # Double tree fraction, and increase veg fraction.
    nday = 31
    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', nday=nday,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=nday,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    # override model1
    model1.blddensity = 0.5
    model1.treecover = 0.125  # 50% tree
    model1.grasscover = 0.125
    # model1.vegcover = 0.25

    # override model2
    model2.blddensity = 0.5
    model2.treecover = 0.25  # 200% tree
    model2.grasscover = 0.125
    # model1.vegcover = 0.375

    model1.generate()
    model2.generate()

    # check vegcover
    assert model1.vegcover == pytest.approx(0.25, 1e-10)
    assert model1.UCM.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.vegcover == pytest.approx(0.375, 1e-10)
    assert model2.UCM.vegcover == pytest.approx(0.375, 1e-10)

    assert model1.road.vegcoverage == pytest.approx(0.5, 1e-10)
    assert model2.road.vegcoverage == pytest.approx(0.75, 1e-10)

    # check tree coverage
    assert model1.UCM.roadShad == pytest.approx(0.25, 1e-10)
    assert model2.UCM.roadShad == pytest.approx(0.5, 1e-10)

    # run models for 7 days starting from Jan 1
    model1.simulate()
    model2.simulate()

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]

    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)

    # confirm if increasing trees result in lower temps
    assert mean_temps1 > mean_temps2


def test_veg_drybulb():
    """Test if drybulb decreases when grasscover increases."""

    # Double grass fraction, and increase veg fraction.
    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    # override model1
    model1.blddensity = 0.5
    model1.treecover = 0.125  # 50% tree
    model1.grasscover = 0.125  # TODO: make into grasscover
    # model1.vegcover = 0.25

    # override model2
    model2.blddensity = 0.5
    model2.treecover = 0.125  # 200% tree
    model2.grasscover = 0.25  # TODO: make the grasscover
    # model1.vegcover = 0.375

    model1.generate()
    model2.generate()

    # check vegcover
    assert model1.vegcover == pytest.approx(0.25, 1e-10)
    assert model1.UCM.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.vegcover == pytest.approx(0.375, 1e-10)
    assert model2.UCM.vegcover == pytest.approx(0.375, 1e-10)

    assert model1.road.vegcoverage == pytest.approx(0.5, 1e-10)
    assert model2.road.vegcoverage == pytest.approx(0.75, 1e-10)

    # check tree coverage
    assert model1.UCM.roadShad == pytest.approx(0.25, 1e-10)
    assert model2.UCM.roadShad == pytest.approx(0.25, 1e-10)

    # run models for 7 days starting from Jan 1
    model1.simulate()
    model2.simulate()

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]

    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)

    # confirm if increasing grass result in lower temps
    assert mean_temps1 > mean_temps2


def test_tree_only_drybulb():
    """Test drybulb impact of treecover change but vegcover is constant.

    Note: while counterintuitive that increasing tree coverage increases dry bulb,
    this is primarily due to the tree canopy blocking long-wave radiation.
    """

    # Double tree fraction, and subtract grass fraction.
    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, albveg=0.5, treecover=0, grasscover=0,
        epw_path=EPW_PATH)

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, albveg=0.5, treecover=0, grasscover=0,
        epw_path=EPW_PATH)

    # override model1
    model1.blddensity = 0.5
    model1.treecover = 0.125  # 50% tree
    model1.grasscover = 0.125
    model1.latgrss = 0.6
    model1.lattree = 0.6
    # model1.vegcover = 0.25

    # override model2
    model2.blddensity = 0.5
    model2.treecover = 0.25  # 200% tree
    model2.grasscover = 0.0
    model2.latgrss = 0.6
    model2.lattree = 0.6
    # model2.vegcover = 0.25

    model1.generate()
    model2.generate()

    # check vegcover
    assert model1.vegcover == pytest.approx(0.25, 1e-10)
    assert model1.UCM.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.UCM.vegcover == pytest.approx(0.25, 1e-10)

    assert model1.road.vegcoverage == pytest.approx(0.5, 1e-10)
    assert model2.road.vegcoverage == pytest.approx(0.5, 1e-10)

    # check tree coverage
    assert model1.UCM.roadShad == pytest.approx(0.25, 1e-10)
    assert model2.UCM.roadShad == pytest.approx(0.5, 1e-10)

    # run models for 7 days starting from Jan 1
    model1.simulate()
    model2.simulate()

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]

    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)

    # increasing trees result in higher temps due to long-wave radiation
    # print(mean_temps1, mean_temps2)
    assert mean_temps1 < mean_temps2


def test_veg_only_drybulb():
    """Test if drybulb decreases when grasscover increases but vegcover is constant."""

    # Double grass fraction, and maintain veg fraction.
    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=7,
        vegstart=1, vegend=12, albveg=0.5, treecover=0, grasscover=0,
        epw_path=EPW_PATH)

    # override model1
    model1.blddensity = 0.5
    model1.treecover = 0.125  # 50% tree
    model1.grasscover = 0.125
    model1.latgrss = 0.4
    model1.lattree = 0.6
    # model1.vegcover = 0.25

    # override model2
    model2.blddensity = 0.5
    model2.treecover = 0.0  # 200% tree
    model2.grasscover = 0.25
    model2.latgrss = 0.4
    model2.lattree = 0.6
    # model1.vegcover = 0.25

    model1.generate()
    model2.generate()

    # check vegcover
    assert model1.vegcover == pytest.approx(0.25, 1e-10)
    assert model1.UCM.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.vegcover == pytest.approx(0.25, 1e-10)
    assert model2.UCM.vegcover == pytest.approx(0.25, 1e-10)

    assert model1.road.vegcoverage == pytest.approx(0.5, 1e-10)
    assert model2.road.vegcoverage == pytest.approx(0.5, 1e-10)

    # check tree coverage
    assert model1.UCM.roadShad == pytest.approx(0.25, 1e-10)
    assert model2.UCM.roadShad == pytest.approx(0.00, 1e-10)

    # run models for 7 days starting from Jan 1
    model1.simulate()
    model2.simulate()

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]

    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)

    # confirm if increasing grass result in lower temps
    assert mean_temps1 > mean_temps2


def test_tree_sens_heat():
    """Test treeSensHeat."""

    model = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=3,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    # override model1
    model.blddensity = 0.5
    model.treecover = 0.1  # 20% tree
    model.grasscover = 0.4
    # model1.vegcover = 0.5

    # run models for 3 days starting from Jan 1
    model.generate()
    model.simulate()

    # check tree coverage
    assert model.UCM.roadShad == pytest.approx(0.2, 1e-10)

    hr = (24 * 2) + 12  # afternoon on last day

    # treeSensHeat is just Solar on road, even when fraction is 0
    sol = model.UCMData[hr].SolRecRoad
    sens_chk = sol * 0.4 * 0.5 * 0.1  # sensFrac*alb*treeFrac = tree sens
    sens_chk += sol * 0.6 * 0.5 * 0.4  # sensFrac*alb*grassFrac = grass sens
    # Note that sensible heat accounts for fraction of tree/grass relative to
    # entire urban area, not relative to road, since SolRecRoad doesn't account
    # for road fraction (just road view factor)
    assert model.UCMData[hr].treeSensHeat == pytest.approx(sens_chk, abs=1e-10)

    # with zero coerage
    model = UWG.from_param_args(
        bldheight=10, blddensity=0, vertohor=0.8, zone='1A', nday=3,
        vegstart=1, vegend=12, latgrss=0.4, lattree=0.6, albveg=0.5,
        treecover=0, grasscover=0, epw_path=EPW_PATH)

    # override model1
    model.blddensity = 0.5
    model.treecover = 0
    model.grasscover = 0  # TODO: make into grasscover
    # model1.vegcover = 0

    # run models for 3 days starting from Jan 1
    model.generate()
    model.simulate()

    # check tree coverage
    assert model.UCM.roadShad == pytest.approx(0.0, 1e-10)

    hr = (24 * 2) + 12  # afternoon on last day

    # treeSensHeat is just Solar on road, even when fraction is 0
    sol = model.UCMData[hr].SolRecRoad
    assert model.UCMData[hr].treeSensHeat == pytest.approx(0, abs=1e-10)


def test_veg_element():
    """Test veg Element calcs"""

    model = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', nday=1,
        treecover=0.125, grasscover=0.125, epw_path=EPW_PATH)

    model.vegstart = 1
    model.vegend = 12
    model.lattree = 0.5
    model.albveg = 0.5
    model.latgrss = 0.5
    model.lattree = 0.5
    model.albroad = 0.25

    # simulate
    model.generate()
    model.simulate()

    assert model.vegcover == pytest.approx(0.25, 1e-10)
    assert model.road.vegcoverage == pytest.approx(0.5, 1e-10)
    assert model.UCM.vegcover == pytest.approx(model.vegcover, 1e-10)

    # Summer, veg
    solRec = 10
    vegcoverage = 0.5
    albedo = 0.25
    grassFLat = 0.5
    vegAlbedo = 0.5

    solAbs = (1.0 - vegcoverage) * (1.0 - albedo) * solRec
    solAbs += vegcoverage * (1.0 - vegAlbedo) * solRec
    vegSens = vegcoverage * (1.0 - vegAlbedo) * solRec * (1.0 - grassFLat)

    solref = (1.0 - vegcoverage) * (albedo) * solRec
    solref += vegcoverage * (vegAlbedo) * solRec

    # check sensible & net heat flux
    sens = vegSens  # + aeroCond * (layerTemp[0] - tempRef)
    flux = -sens + solAbs  # + infra - lat  # [W m-2]

    # print to understand interactions
    # print('solrefl:', solref)
    # print('% solAbs:', (0.5 * 0.75) + (0.5 * 0.75))
    # print('solAbs: ', solAbs)
    # print('% vegSens:', 0.5 * 0.5 * 0.5)
    # print('vegSens: ', vegSens)
    # print('sens: ', sens)
    # print('flux: ', flux)

    UCM = model.UCMData[12]  # afternoon
    sol = model.UCMData[12].SolRecRoad
    treeSensHeat = sol * 0.5 * 0.5 * model.vegcover
    assert treeSensHeat == pytest.approx(UCM.treeSensHeat, 1e-10)


def test_veg_roof_sens():
    """Test veg roof calcs.
    """

    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', month=6,
        nday=31, treecover=0.0, grasscover=0.0, vegstart=1, vegend=12, dtsim=300,
        epw_path=EPW_PATH)
    model1.vegroof = 0.1
    model1.latgrss = 0.4

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0.5, vertohor=0.8, zone='1A', month=6,
        nday=31, treecover=0.0, grasscover=0.0, vegstart=1, vegend=12, dtsim=300,
        epw_path=EPW_PATH)
    model2.vegroof = 1
    model2.latgrss = 0.4

    # generate
    model1.generate()
    model2.generate()

    print(model1.BEM)

    # check fractions
    for i in range(2):
        assert model1.BEM[i].roof.vegcoverage == pytest.approx(0.1, 1e-10)
        assert model2.BEM[i].roof.vegcoverage == pytest.approx(1, 1e-10)

    # simulate
    model1.simulate()
    model2.simulate()

    # Roof vegetation sensible heat is greater with more vegetation, but convection
    # here mitigates that. This only works b/c latgrss is very low so that veg
    # sens heat is > then convection. For checking only, no test.
    # qroof1 = [ucm.Q_roof for ucm in model1.UCMData]
    # qroof2 = [ucm.Q_roof for ucm in model2.UCMData]
    # mean_qroof1, mean_qroof2 = sum(qroof1) / len(qroof1), sum(qroof2) / len(qroof2)
    # print(mean_qroof1, mean_qroof2)

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]

    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)

    # Increasing roof vegetation results in slightly lower temps, as long as
    # latgrss is very low, which means sens heat is > then convection.
    # print(mean_temps1, mean_temps2)
    assert mean_temps1 > mean_temps2


def test_grass_sens():
    """Test changing grass sens makes a difference.

    This tests if changing grass sensible heat fraction on tree sensible heat,
    urban sensible heat, and drybulb.

    While rural grass sensible heat is accounted for in UWG, the code from
    the UWG_Matlab doesn't account for grass sensible heat on the urban
    road.

    To verify that the UHI doesn't change with changes to the sensible heat
    fraction of grass, comment out the following code to eliminate the
    impact of the rural road grass, and revert treeSensHeat to original
    UWG_Matlab code:
        At UBLDef.ublmodel  // set heatDif = max(self.sensHeat - rural.sens, 0) to 0
        At RSMDef.vdm  //set rural.sens to 1.0
    """

    model1 = UWG.from_param_args(
        bldheight=10, blddensity=0.0, vertohor=0.8, grasscover=0.1, treecover=0.1,
        zone='1A', nday=31, vegstart=1, vegend=12, dtsim=600, epw_path=EPW_PATH)
    # 100% of open space is grass with 0.9 sensible heat
    model1.vegroof = 0.0
    model1.blddensity = 0.5
    model1.treecover = 0.0
    model1.grasscover = 0.4
    model1.latgrss = 0.3

    model2 = UWG.from_param_args(
        bldheight=10, blddensity=0.0, vertohor=0.8, grasscover=0.1, treecover=0.1,
        zone='1A', nday=31, vegstart=1, vegend=12, dtsim=600, epw_path=EPW_PATH)
    # 100% of open space is grass with 0.1 sensible heat
    model2.vegroof = 0.0
    model2.blddensity = 0.5
    model2.treecover = 0.0
    model2.grasscover = 0.4
    model2.latgrss = 0.7

    # model1 should have produce more vegetation sensible heat.

    # generate & simulate
    model1.generate()
    model1.simulate()
    model2.generate()
    model2.simulate()

    # confirm tree/veg coverage has more sensible veg contribution with more sens factor.
    # The solar blocked by veg will be more then sensible heat fraction, so net effect
    # is less heat to canopy.
    qtree1 = [ucm.treeSensHeat for ucm in model1.UCMData]
    qtree2 = [ucm.treeSensHeat for ucm in model2.UCMData]
    mean_qtree1, mean_qtree2 = sum(
        qtree1) / len(qtree1), sum(qtree2) / len(qtree2)
    # print(mean_qtree1, mean_qtree2)
    assert mean_qtree1 > mean_qtree2

    # confirm canopy sens heat
    qsens1 = [ucm.sensHeat for ucm in model1.UCMData]
    qsens2 = [ucm.sensHeat for ucm in model2.UCMData]
    mean_qsens1, mean_qsens2 = sum(
        qsens1) / len(qsens1), sum(qsens2) / len(qsens2)
    # print(mean_qsens1, mean_qsens2)
    assert mean_qsens1 > mean_qsens2

    # calc drybulb mean
    temps1 = [ucm.canTemp - 273.15 for ucm in model1.UCMData]
    temps2 = [ucm.canTemp - 273.15 for ucm in model2.UCMData]
    mean_temps1, mean_temps2 = sum(
        temps1) / len(temps1), sum(temps2) / len(temps2)
    # print(mean_temps1, mean_temps2)
    assert mean_temps1 > mean_temps2  # higher latent fraction makes cooler temps



"""Test for uwg.py"""
import os
import pytest
from .test_base import auto_setup_uwg

DIR_CURR = os.path.abspath(os.path.dirname(__file__))
DEFAULT_EPW_PATH = os.path.join(
    DIR_CURR, 'epw', 'SGP_Singapore.486980_IWEC.epw')
DEFAULT_PARAM_PATH = os.path.join(
    DIR_CURR, 'parameters', 'initialize_singapore.uwg')


def test_required_inputs_wrong_type():
    """Test that we can catch wrong types, list lengths etc."""

    testuwg = auto_setup_uwg(DEFAULT_EPW_PATH, DEFAULT_PARAM_PATH)

    with pytest.raises(AssertionError):
        testuwg.h_temp = -6
    testuwg.h_temp = 2.0

    with pytest.raises(AssertionError):
        testuwg.windmin = -100
    testuwg.windmin = 2.0

    # assert type(self.Day) == float or type(self.Day) == int
    with pytest.raises(ValueError):
        testuwg.day = 'a'

    testuwg.day = 2
    testuwg.day = 4.0

    # assert isinstance(self.SchTraffic, list)
    with pytest.raises(AssertionError):
        testuwg.schtraffic = 76

    testuwg.schtraffic = [[0.1 for i in range(24)]] * 3  # length 3 list

    with pytest.raises(AssertionError):
        testuwg.bld = 4543

    testuwg.bld = [['test', 'pre80', 0.1], ['test2', 'pre80', 0.9]]



"""Tests for urbflux functions."""

import pytest
from .test_base import auto_setup_uwg, setup_open_matlab_ref, calculate_tolerance


def test_urbflux_unit():
    """Test for urbflux."""

    testuwg = auto_setup_uwg()

    # Test Jan 1 (winter, no vegetation coverage)
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 1

    # set_input
    testuwg.generate()

    # In order to avoid integration effects. Test only first time step
    # Subtract timestep to stop at 300 sec
    testuwg.simTime.nt -= (23 * 12 + 11)

    # Run simulation
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    # Calculated values
    uwg_python_val = [
        # building surface properties
        testuwg.BEM[1].roof.emissivity,        # roof emissivity
        testuwg.BEM[1].roof.layerTemp[0],      # roof surface temp (K)
        testuwg.BEM[1].roof.infra,             # roof net longwave radiation (W m-2)
        testuwg.BEM[1].mass.layerTemp[-1],
        testuwg.BEM[1].wall.emissivity,        # wall emissivity
        testuwg.BEM[1].wall.layerTemp[0],      # wall surface temp (K)
        testuwg.BEM[1].wall.infra,             # wall net longwave radiation (W m-2)
        testuwg.BEM[1].wall.layerTemp[-1],     # wall interior surface temp (K)
        # urban canyon properties
        testuwg.UCM.wallTemp,                  # canyon wall temp (K)
        testuwg.UCM.roofTemp,                  # canyon roof temp (K)
        testuwg.road.infra,                    # road net longwave radiation (W m-2)
        testuwg.UCM.roadTemp,                  # canyon road temperature (W m-2)
        testuwg.UCM.latHeat,                   # canyon latent heat (W m-2)
        testuwg.UBL.advHeat,                   # boundary layer advective heat (W m-2)
        testuwg.UCM.ustar,                     # canyon friction velocity (m s-1)
        testuwg.UCM.ustarMod,                  # modified friction velocity (m s-1)
        testuwg.UCM.uExch,                     # exchange velocity (m s-1)
        testuwg.UCM.canWind,                   # canyon wind velocity (m s-1)
        testuwg.UCM.turbU,                     # canyon turbulent velocity (m s-1)
        testuwg.UCM.turbV,                     # canyon turbulent velocity (m s-1)
        testuwg.UCM.turbW,                     # canyon turbulent velocity (m s-1)
        testuwg.UCM.windProf[-1]               # canyon wind profile
    ]

    # UBL.advHeat = Boundary layer advective heat rounding error
    # Past 10^-9 it's basically sero, but for accuracy's sake apply
    # Kahan's summation algorithm
    # -6.87022037029e-11  ==  -3.43511e-11 << w/o Kahan summation algo
    # -6.01144282401e-11  ==  -3.43511e-11 << w Kahan summation algo

    uwg_matlab_val = \
        setup_open_matlab_ref('matlab_urbflux', 'matlab_ref_urbflux_unit.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        if uwg_matlab_val[i] == '\n':
            uwg_matlab_val[i] = None
            assert uwg_python_val[i] == uwg_matlab_val[i], 'error at index={}'.format(i)
            continue

        elif i == 13:
            tol = 10.  # hardcoding tolerance b/c advheat rounding errors

        else:
            tol = calculate_tolerance(uwg_python_val[i], 15.0)

        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            'error at index={}'.format(i)



"""Class for Urban Canopy Model."""
from __future__ import division

try:
    range = xrange
except NameError:
    pass

from math import sqrt, pow
import copy


class UCMDef(object):
    """Definition of Urban Canopy - Building Energy Model Class

    Args:
        bldHeight: Average building height (m).
        bldDensity: Horizontal building density (footprint/total_area).
        verToHor: Vertical-to-horizontal urban area ratio (facade area/urban area).
        treeCoverage: Horizontal tree density (footprint).
        sensAnthrop: Sensible anthropogenic heat (other than from buildings) (W m-2).
        latAnthrop: Latent anthropogenic heat (other than from buildings) (W m-2).
        initialTemp: Canyon air temperature (db) (K).
        initialHum: Canyon specific humidity (kg kg-1).
        initialWind: Urban canyon wind velocity (m s-1).
        parameter: Param object.
        r_glaze: Area-weighted average of glazing ratio from urban building stock.
        SHGC: Area-weighted average of SHGC from urban building stock.
        alb_wall: Area-weighted average of wall albedo from urban building stock.
        road: Road element class (moved from BEM).

    Properties
        * road -- Road element class (moved from BEM)
        * # Urban Canyon Parameters
        * bldHeight -- average building height (m)
        * bldDensity -- horizontal building density (footprint)
        * verToHor -- vertical-to-horizontal urban area ratio (facade area/urban area)
        * treeCoverage -- horizontal tree density (footprint)
        * sensAnthrop -- sensible anthropogenic heat (other than from buildings) (W m-2)
        * latAnthrop -- latent anthropogenic heat (other than from buildings) (W m-2)
        * z0u -- urban roughness length (m)
        * l_disp -- urban displacement length (m)
        * roadShad -- shadowing of roads
        * canWidth -- canyon width (m)
        * bldWidth -- bld width (m)
        * canAspect -- canyon aspect ratio
        * roadConf -- road-sky configuration factors (sky view factor)
        * alb_wall -- average wall albedo
        * wallConf -- wall-sky configuration factors (sky view factor)
        * VFwallroad -- wall-road view factor
        * VFroadwall -- road-wall view factor
        * facArea -- facade area (m2)
        * roadArea -- road area (m2)
        * roofArea -- roof area (m2) (also building area)
        * facAbsor -- average facade absortivity
        * roadAbsor -- average road absortivity
        * h_mix -- waste heat mix into canyon ratio
        * # Urban Canyon Variables
        * canTemp -- canyon air temperature (db) (K)
        * Tdp -- dew point temperature
        * Twb -- wetbulb temperature
        * canHum -- canyon specific humidity (kg kg-1)
        * canRHum -- canyon relative humidity ()
        * canWind -- urban canyon wind velocity (m s-1)
        * turbU -- canyon turbulent velocities (m s-1)
        * turbV --  canyon turbulent velocities (m s-1)
        * turbW -- canyon turbulent velocities (m s-1)
        * ublTemp -- urban boundary layer temperature (K)
        * ublTempdx -- urban boundary layer temperature discretization (K)
        * ublWind -- urban boundary layer wind velocity (m s-1)
        * ustar -- friction velocity (m s-1)
        * ustarMod -- modified friction velocity (m s-1)
        * uExch -- exchange velocity (m s-1)
        * treeLatHeat -- latent heat from trees (W m-2)
        * treeSensHeat -- sensible heat from trees (W m-2)
        * sensHeat -- urban sensible heat (W m-2)
        * latHeat -- urban latent heat (W m-2)
        * windProf -- urban wind profile
        * Q_roof -- sensible heat flux from building roof (convective)
        * Q_wall -- sensible heat flux from building wall (convective)
        * Q_window -- sensible heat flux from building window (via U-factor)
        * Q_road -- sensible heat flux from road (convective)
        * Q_hvac -- sensible heat flux from HVAC waste
        * Q_traffic -- sensible heat flux from traffic (net)
        * Q_ubl -- Convective heat exchange with UBL layer
        * Q_vent -- Convective heat exchange from ventilation/infiltration
        * SolRecWall -- Solar received by wall
        * SolRecRoof -- Solar received by roof
        * SolRecRoad -- Solar received by road
        * roadTemp -- average road temperature (K)
        * roofTemp -- average roof temperature (K)
        * wallTemp -- average wall temperature (K)
        * ElecTotal -- Total Electricity consumption of urban area
        * GasTotal -- Total Gas consumption of the urban area
    """

    def __init__(self, bldHeight, bldDensity, verToHor, treeCoverage, sensAnthrop,
                 latAnthrop, initialTemp, initialHum, initialWind, parameter,
                 r_glaze, SHGC, alb_wall, road):

        self.road = road
        self.bldHeight = bldHeight
        self.verToHor = verToHor
        self.bldDensity = bldDensity
        self.treeCoverage = treeCoverage
        self.vegcover = (1 - self.bldDensity) * self.road.vegcoverage
        self.sensAnthrop = sensAnthrop
        self.latAnthrop = latAnthrop
        self.roadShad = min(treeCoverage / (1 - bldDensity), 1)
        # UWG_Matlab assumes bld_area = square, so sqrt(bld_area) = side length
        # bld width: (side length) derived from bldDensity and verToHor (m)
        self.bldWidth = 4 * bldHeight * bldDensity / verToHor
        # urban area width == sqrt(bldDensity) == \
        # ratio of bld footprint_width / urban area footprint
        d = self.bldWidth / (sqrt(bldDensity))
        # canyon width (m) = urban area width - building width
        self.canWidth = d - self.bldWidth
        self.canAspect = bldHeight / self.canWidth
        self.roadConf = pow(pow(self.canAspect, 2) + 1, 0.5) - self.canAspect
        self.wallConf = (
            0.5 * (self.canAspect + 1 - pow(pow(self.canAspect, 2) + 1, 0.5)) /
            self.canAspect)
        self.facArea = 4 * self.bldWidth * bldHeight
        self.roadArea = d * d - pow(self.bldWidth, 2)
        self.roofArea = pow(self.bldWidth, 2)
        self.canTemp = initialTemp
        self.roadTemp = initialTemp
        self.canHum = initialHum
        self.ublWind = max(initialWind, parameter.windMin)
        self.canWind = initialWind
        self.ustar = 0.1 * initialWind
        self.ustarMod = 0.1 * initialWind

        # Calculate z0u = urban roughness length (m)
        frontDens = verToHor / 4.  # density of just street facing facade
        if frontDens < 0.15:
            self.z0u = frontDens * self.bldHeight
        else:
            self.z0u = 0.15 * self.bldHeight

        # Calculate l_dsp = urban displacement length (m)
        if frontDens < 0.05:
            self.l_disp = 3 * frontDens * self.bldHeight
        elif frontDens < 0.15:
            self.l_disp = (0.15 + 5.5 * (frontDens - 0.05)) * self.bldHeight
        elif frontDens < 1:
            self.l_disp = (0.7 + 0.35 * (frontDens - 0.15)) * self.bldHeight
        else:
            self.l_disp = 0.5 * self.bldHeight

        self.alb_wall = alb_wall
        self.facAbsor = (1 - r_glaze) * (1 - alb_wall) + r_glaze * (1 - 0.75 * SHGC)
        self.roadAbsor = (1 - road.vegcoverage) * (1 - road.albedo)
        self.sensHeat = 0.0

        # Variables set in urbflux function
        self.latHeat = None
        self.windProf = []
        self.canRHum = None
        self.Tdp = None

    def UCModel(self, BEM, T_ubl, forc, parameter):
        """Calculate the urban canyon temperature per The uwg (2012) Eq. 10."""

        # air density
        dens = \
            forc.pres / (1000 * 0.287042 * self.canTemp * (1. + 1.607858 * self.canHum))
        dens_ubl = forc.pres / (1000 * 0.287042 * T_ubl * (1. + 1.607858 * forc.hum))
        Cp_air = parameter.cp

        self.Q_wall = 0.
        self.Q_window = 0.
        self.Q_road = 0.
        self.Q_hvac = 0.
        self.Q_traffic = 0.
        self.Q_vent = 0.
        self.Q_ubl = 0.
        self.ElecTotal = 0.
        self.GasTotal = 0.
        self.roofTemp = 0.
        self.wallTemp = 0.

        # Road to Canyon
        T_road = self.road.layerTemp[0]
        h_conv = self.road.aeroCond
        H1 = T_road * h_conv * self.roadArea  # Heat (Sens) from road surface
        H2 = h_conv * self.roadArea

        H1 = H1 + T_ubl * self.roadArea * self.uExch * Cp_air * dens_ubl  # Heat from UBL
        H2 = H2 + self.roadArea * self.uExch * Cp_air * dens_ubl
        # W = m2 * W/m2
        Q = (self.roofArea + self.roadArea) * (self.sensAnthrop + self.treeSensHeat)

        # Building energy output to canyon, in terms of absolute (total) values
        for j in range(len(BEM)):
            # Re-naming variable for readability
            building = BEM[j].building
            wall = BEM[j].wall
            T_indoor = building.indoor_temp
            T_wall = wall.layerTemp[0]
            R_glazing = building.glazing_ratio
            A_wall = (1. - R_glazing) * self.facArea
            A_window = R_glazing*self.facArea
            U_window = building.u_value

            H1 = H1 + BEM[j].frac*(
                T_indoor * A_window * U_window +  # window U
                T_wall * A_wall * h_conv +  # wall conv
                (T_indoor * self.roofArea * BEM[j].building.vent *
                 BEM[j].building.nFloor * Cp_air * dens) +  # Vent
                (T_indoor * self.roofArea * BEM[j].building.infil * self.bldHeight /
                 3600.0 * Cp_air * dens))  # Infil

            H2 = H2 + BEM[j].frac * (
                A_window * U_window +  # window U
                A_wall * h_conv +  # wall conv
                (self.roofArea * BEM[j].building.vent * BEM[j].building.nFloor *
                 Cp_air * dens) +  # Vent
                (self.roofArea * BEM[j].building.infil * self.bldHeight / 3600.0 *
                 Cp_air * dens))  # Infil

            Q = Q + BEM[j].frac * (
                self.roofArea * building.sensWaste * self.h_mix +  # HVAC waste heat
                # heat that didn't make it to inside
                A_window * BEM[j].wall.solRec * (1.0 - BEM[j].building.shgc))

            self.wallTemp = self.wallTemp + BEM[j].frac * T_wall
            self.roofTemp = self.roofTemp + BEM[j].frac * BEM[j].roof.layerTemp[0]
            self.Q_ubl = (self.Q_ubl + BEM[j].frac * self.bldDensity *
                          (BEM[j].roof.sens + BEM[j].building.sensWaste *
                          (1. - self.h_mix)))  # Changed by Jiachen Mao in March 2017

        # Solve for canyon temperature
        self.canTemp = (H1 + Q) / H2

        # Heat flux based per m^2 of urban area
        # Sensible heat from road (W/m^2 of urban area)
        self.Q_road = h_conv * (T_road - self.canTemp) * (1. - self.bldDensity)
        self.Q_ubl = (
            self.Q_ubl + self.uExch * Cp_air * dens * (self.canTemp - T_ubl) *
            (1. - self.bldDensity))
        self.Q_wall = h_conv * (self.wallTemp - self.canTemp) * self.verToHor
        self.Q_traffic = self.sensAnthrop

        # Building energy output to canyon, per m^2 of urban area
        T_can = copy.copy(self.canTemp)
        for j in range(len(BEM)):
            # ventilation volume per m^2 of building
            V_vent = BEM[j].building.vent * BEM[j].building.nFloor
            V_infil = BEM[j].building.infil * self.bldHeight / 3600.0
            T_indoor = BEM[j].building.indoor_temp
            U_window = BEM[j].building.u_value  # Added by Jiachen Mao in 03/17
            R_glazing = BEM[j].building.glazing_ratio  # Changed by Jiachen Mao in 03/17

            self.Q_window = (self.Q_window + BEM[j].frac * self.verToHor * R_glazing *
                             U_window * (T_indoor - T_can))
            self.Q_window = (self.Q_window + BEM[j].frac * self.verToHor * R_glazing *
                             BEM[j].wall.solRec * (1. - BEM[j].building.shgc))
            self.Q_vent = (self.Q_vent + BEM[j].frac * self.bldDensity * Cp_air * dens *
                           (V_vent + V_infil) * (T_indoor - T_can))
            self.Q_hvac = (self.Q_hvac + BEM[j].frac * self.bldDensity *
                           BEM[j].building.sensWaste * self.h_mix)

            self.Q_roof = self.Q_roof + BEM[j].frac * self.bldDensity * BEM[j].roof.sens

            # Total Electrical & Gas power in MW
            self.ElecTotal = \
                self.ElecTotal + BEM[j].fl_area * BEM[j].building.ElecTotal / 1.e6
            self.GasTotal = \
                self.GasTotal + BEM[j].fl_area * BEM[j].building.GasTotal / 1.e6

        # ------------------------------------------------------------------------------
        # Sensible Heat
        # Note 1: In the current uwg code, latent heat from evapotranspiration, stagnant
        # water, or anthropogenic sources is not modeled due to the difficulty of
        # validation, and lack of reliability of precipitation data from EPW files.
        # Note 2: Changed sensHeat to multiply treeSensHeat by treeCoverage fraction.
        # Since treeSensHeat calculated in solarcalcs does not multiply sensheat by tree
        # fraction. For example, if there were 0 trees, this value should be 0.
        # ------------------------------------------------------------------------------
        self.sensHeat = (
            self.Q_wall + self.Q_road + self.Q_vent + self.Q_window + self.Q_hvac +
            self.Q_traffic + self.treeSensHeat + self.Q_roof)

        # Error checking
        if self.canTemp > 350. or self.canTemp < 250:
            raise Exception('Got canyon temperature at {} C. Something obviously went '
                            'wrong (UCMDef.py).'.format(self.canTemp - 273.15))

    def __repr__(self):
        return 'UCMDef,\n verToHor: {}\n bldDensity: {}\n bldHeight: {}\n canWidth: ' \
            '{}\n canAspect: {}\n facArea: {}\n roofArea: {}'.format(
                self.verToHor, self.bldDensity, self.bldHeight, self.canWidth,
                self.canAspect, self.facArea, self.roofArea)



"""Functions for calculating canyon infrared radiation."""

SIGMA = 5.67e-8  # Stephen-Boltzman const


def infracalcs(UCM, forc, e_road, e_wall, T_road, T_wall):
    """Calculate infrared radiation surface flux on canyon wall and road.

    Args:
        UCM: UCMDef object.
        forc: Forcing object.
        e_road: Number for road emissivity.
        e_wall: Number for wall emissivity.
        T_road: Number for road surface temperature (K).
        T_wall: Number for wall surface temperature (K).

    Returns:
        Tuple of two values

        - infra_road: Road infrared radiation (W)

        - infra_wall: Wall infrared radiation (W)
    """
    # road_wall_conf: configuration factor (view factor) for road to wall
    road_wall_conf = (1. - UCM.roadConf)
    # wall_road_conf: wall to road VF same as wall-sky configuration factors
    # (sky view factor)
    wall_road_conf = UCM.wallConf

    # Calculate radiation of unshaded road, accounting for radiation exchange from wall
    _road_rad = \
        e_road * UCM.roadConf * (1. - UCM.roadShad) * (forc.infra - SIGMA * T_road ** 4.)
    _wall_to_road_rad = \
        (1. - UCM.roadShad) * e_wall * e_road * SIGMA * road_wall_conf * (T_wall ** 4. - T_road ** 4.)

    infra_road = _road_rad + _wall_to_road_rad

    # Calculate radiation of wall, accounting for radiation exchange from unshaded road
    _road_rad = e_wall * UCM.wallConf * (forc.infra - SIGMA * T_wall ** 4.)
    _road_to_wall_rad = \
        (1. - UCM.roadShad) * e_wall * e_road * SIGMA * wall_road_conf * (T_road ** 4. - T_wall ** 4.)

    infra_wall = _road_rad + _road_to_wall_rad

    return infra_road, infra_wall



"""Tests for Param object."""

import os
from uwg import Param, UWG


def test_param_init():
    """Param init test."""

    epw_path = \
        os.path.join(os.path.dirname(__file__), 'epw',
                     'SGP_Singapore.486980_IWEC.epw')
    testuwg = UWG.from_param_args(
        10, 0.5, 0.5, 0.1, 0.1, '1A', epw_path=epw_path)

    nightStart = 18.
    nightEnd = 8.
    maxdx = 250.

    geoparam = Param(
        testuwg.h_ubl1, testuwg.h_ubl2, testuwg.h_ref, testuwg.h_temp, testuwg.h_wind,
        testuwg.c_circ, testuwg.maxday, testuwg.maxnight, testuwg.lattree,
        testuwg.latgrss, testuwg.albveg, testuwg.vegstart, testuwg.vegend, nightStart,
        nightEnd, testuwg.windmin, testuwg.WGMAX, testuwg.c_exch, maxdx, testuwg.G,
        testuwg.CP, testuwg.VK, testuwg.R, testuwg.RV, testuwg.LV, 3.14, testuwg.SIGMA,
        testuwg.WATERDENS, testuwg.LVTT, testuwg.TT, testuwg.ESTT, testuwg.CL,
        testuwg.CPV, testuwg.B, testuwg.CM, testuwg.COLBURN)

    geoparam.__repr__()



"""Test the matlab generated epw file against the python generated epw files."""

import pytest
import os

from .test_base import auto_setup_uwg, calculate_tolerance, set_input_manually, \
    MATLAB_DIR, TEST_DIR
import uwg


def compare_epw(testuwg, matlab_fname, precision=10.0):
    # shorten some variable names
    ti = testuwg.simTime.timeInitial
    tf = testuwg.simTime.timeFinal

    # Make weather files for testing
    matlab_path_name = os.path.join(MATLAB_DIR, 'matlab_output', matlab_fname)

    pywtr = uwg.weather.Weather(testuwg.new_epw_path, ti, tf)
    matwtr = uwg.weather.Weather(matlab_path_name, ti, tf)

    assert len(pywtr.staTemp) == pytest.approx(len(matwtr.staTemp), abs=1e-15)

    for i in range(0, len(pywtr.staTemp)):

        # Check dry bulb [K]
        tol = calculate_tolerance(pywtr.staTemp[i], precision)
        assert pywtr.staTemp[i] == \
            pytest.approx(matwtr.staTemp[i], abs=tol), 'error at index={}'.format(i)

        # Check dew point [K]
        tol = calculate_tolerance(pywtr.staTdp[i], precision)
        assert pywtr.staTdp[i] == \
            pytest.approx(matwtr.staTdp[i], abs=tol), 'error at index={}'.format(i)

        # Check relative humidity [%]
        tol = calculate_tolerance(pywtr.staRhum[i], precision)
        assert pywtr.staRhum[i] == \
            pytest.approx(matwtr.staRhum[i], abs=tol), 'error at index={}'.format(i)

        # Check wind speed [m/s]
        tol = calculate_tolerance(pywtr.staUmod[i], precision)
        assert pywtr.staUmod[i] == \
            pytest.approx(matwtr.staUmod[i], abs=tol), 'error at index={}'.format(i)


def test_uwg_output_heatdemand_1_1_0000():
    """Initial conditions
        - night time
        - before vegstart
        - sensHeatDemand
    """
    epw_path = os.path.join(TEST_DIR, 'epw', 'CAN_ON_Toronto.716240_CWEC.epw')
    testuwg = auto_setup_uwg(epw_path=epw_path)

    # Test all year
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 365

    testuwg.generate()
    testuwg.simulate()
    testuwg.write_epw()

    compare_epw(testuwg, 'CAN_ON_Toronto.716240_CWEC_heatdemand_UWG_Matlab.epw')


def test_program_input():

    epw_path = os.path.join(TEST_DIR, 'epw', 'SGP_Singapore.486980_IWEC.epw')
    testuwg = auto_setup_uwg(epw_path=epw_path, param_path=None)

    # Assign manually
    testuwg = set_input_manually(testuwg)

    # main
    testuwg.generate()
    testuwg.simulate()
    testuwg.write_epw()

    # Check some of the inputs

    # Check building parameters
    assert testuwg.BEM[0].building.coolcap == \
           pytest.approx((3525.66904 * 1000.0) / 46320.0, abs=1e-3)
    assert testuwg.BEM[0].building.heat_cap == \
           pytest.approx((2875.97378 * 1000.0) / 46320.0, abs=1e-3)
    assert testuwg.BEM[1].building.coolcap == \
           pytest.approx((252.20895 * 1000.0) / 3135., abs=1e-2)
    assert testuwg.BEM[1].building.heat_cap == \
           pytest.approx((132.396 * 1000.0) / 3135., abs=1e-2)

    # Check that final day of timestep is at correct dayType
    assert testuwg.dayType == 1
    assert testuwg.schtraffic[testuwg.dayType - 1][testuwg.simTime.hourDay] == \
        pytest.approx(0.2, abs=1e-6)

    compare_epw(testuwg, 'SGP_Singapore.486980_IWEC_UWG_Matlab.epw')


def test_program_hybrid_input():
    """Testing inputting with api and with .uwg file."""

    testuwg = auto_setup_uwg()

    # Assign manually
    testuwg = set_input_manually(testuwg)

    # override (from intialize_singapore.uwg)
    testuwg.bldheight = 10
    testuwg.h_mix = 1
    testuwg.blddensity = 0.5
    testuwg.vertohor = 0.8
    testuwg.charlength = 1000
    testuwg.albroad = 0.1
    testuwg.d_road = 0.5
    testuwg.sensanth = 20

    # main
    testuwg._read_epw()
    testuwg._compute_BEM()
    testuwg._compute_input()
    testuwg._hvac_autosize()
    testuwg.simulate()
    testuwg.write_epw()

    # Check some of the inputs

    # Check building parameters
    assert testuwg.BEM[0].building.coolcap == \
           pytest.approx((3525.66904 * 1000.0) / 46320.0, abs=1e-3)
    assert testuwg.BEM[0].building.heat_cap == \
           pytest.approx((2875.97378 * 1000.0) / 46320.0, abs=1e-3)
    assert testuwg.BEM[1].building.coolcap \
        == pytest.approx((252.20895 * 1000.0) / 3135., abs=1e-2)
    assert testuwg.BEM[1].building.heat_cap == \
           pytest.approx((132.396 * 1000.0) / 3135., abs=1e-2)

    # Check that final day of timestep is at correct dayType
    assert testuwg.dayType == 1
    assert testuwg.schtraffic[testuwg.dayType - 1][testuwg.simTime.hourDay] == \
        pytest.approx(0.2, abs=1e-6)

    compare_epw(testuwg, 'SGP_Singapore.486980_IWEC_UWG_Matlab.epw')


def test_uwg_output_beijing():
    """Initial conditions:
        - day time
        - after vegstart
        - sensHeatDemand
    """

    epw_path = os.path.join(TEST_DIR, 'epw', 'CHN_Beijing.Beijing.545110_IWEC.epw')
    param_path = os.path.join(TEST_DIR, 'parameters', 'initialize_beijing.uwg')
    testuwg = auto_setup_uwg(epw_path=epw_path, param_path=param_path)

    # Test all year
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 365

    # main
    testuwg.generate()
    testuwg.simulate()
    testuwg.write_epw()

    compare_epw(testuwg, 'CHN_Beijing.Beijing.545110_IWEC_UWG_Matlab.epw')


def test_uwg_output_cooldemand_6_1_0000():
    """Initial conditions:
        - day time
        - after vegstart
        - sensHeatDemand
    """

    # set up the logger
    epw_path = os.path.join(TEST_DIR, 'epw', 'CAN_ON_Toronto.716240_CWEC.epw')
    testuwg = auto_setup_uwg(epw_path=epw_path)

    # Test 30 days in summer
    testuwg.month = 6
    testuwg.day = 1
    testuwg.nday = 30

    # main
    testuwg.generate()
    testuwg.simulate()
    testuwg.write_epw()

    testuwg = compare_epw(testuwg, 'CAN_ON_Toronto.716240_CWEC_cooldemand_UWG_Matlab.epw')


def test_uwg_output_cooldemand_1_1_0000():
    """Initial conditions:
        - night time
        - before vegstart
        - sensCoolDemand
    """
    testuwg = auto_setup_uwg()

    # Test all year
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 365

    # main
    testuwg.generate()
    testuwg.simulate()
    testuwg.write_epw()

    compare_epw(testuwg, 'SGP_Singapore.486980_IWEC_UWG_Matlab.epw')




"""Element class."""
from __future__ import division

try:
    range = xrange
except NameError:
    pass

import math
from .utilities import is_near_zero, float_in_range, float_positive
from .material import Material

try:
    str = basestring
except NameError:
    pass


class Element(object):
    """Element object defines wall, and roof constructions.

    Args:
        albedo: Number for outer surface albedo.
        emissivity: Number for outer surface emissivity.
        layer_thickness_lst: List of layer thickness in meters.
        material_lst: List of Material objects in Element.
        vegcoverage: Number for fraction of surface vegetation coverage.
        t_init: Element initial temperature [K].
        horizontal: Boolean indicating if Element is horizontal or not (vertical).
        name: Text string for name of Element.

    Properties:
        * albedo -- outer surface albedo
        * emissivity -- outer surface emissivity.
        * layer_thickness_lst -- list of layer thickness [m]
        * material_lst -- material objects in Element.
        * layerThermalCond -- vector of layer thermal conductivities [W m-1 K-1]
        * layerVolHeat -- vector of layer volumetric heat [J m-3 K-1]
        * vegcoverage -- surface grass coverage
        * t_init -- element initial temperature [K].
        * layerTemp -- vector of layer temperatures [K]
        * waterStorage -- thickness of water film [m] for horizontal surfaces only
        * horizontal -- 1-horizontal, 0-vertical
        * solRec -- solar radiation received [W m-2]
        * infra -- net longwave radiation [W m-2]
        * lat -- surface latent heat flux [W m-2]
        * sens -- surface sensible heat flux [W m-2]
        * solAbs -- solar radiation absorbed [W m-2]
        * aeroCond -- convective heat transfer coefficient [-]
        * T_ext -- external surface temperature [K]
        * T_int -- internal surface temperature [K]
        * flux -- external surface heat flux [W m-2]
    """

    def __init__(self, albedo, emissivity, layer_thickness_lst, material_lst,
                 vegcoverage, t_init, horizontal, name):

        assert len(layer_thickness_lst) == len(material_lst), 'The number of layer ' \
            'thickness must match the number of layer materials. Got {} and {}, ' \
            'respectively.'.format(len(layer_thickness_lst), len(material_lst))

        self.albedo = albedo  #
        self.emissivity = emissivity  # outer surface emissivity.
        # list of layer thickness [m]
        self.layer_thickness_lst = layer_thickness_lst
        self.material_lst = material_lst  #
        self.vegcoverage = vegcoverage  #
        self.t_init = t_init  #
        self.horizontal = int(horizontal)  #
        self._name = name

        # layerThermaCond:
        self.layerThermalCond = [
            material.thermalcond for material in material_lst]
        # layerVolHeat:
        self.layerVolHeat = [material.volheat for material in material_lst]
        # layerTemp: #
        self.layerTemp = [t_init] * len(layer_thickness_lst)
        # waterStorage: #
        self.waterStorage = 0
        self.infra = 0  #
        self.sens = 0  #
        self.solRec = 0  #
        self.lat = 0  #
        self.solAbs = 0  #
        self.aeroCond = 0  #
        self.T_ext = 293  #
        self.T_int = 293  #
        self.flux = 0  #

    @property
    def albedo(self):
        """Get or set a value between 0 and 1 for outer surface albedo."""
        return self._albedo

    @albedo.setter
    def albedo(self, value):
        self._albedo = float_positive(value, 'albedo')

    @property
    def emissivity(self):
        """Get or set a value between 0 and 1 for outer surface emissivity."""
        return self._emissivity

    @emissivity.setter
    def emissivity(self, value):
        self._emissivity = float_positive(value, 'emissivity')

    @property
    def layer_thickness_lst(self):
        """Get or set list of thickness in meters of each Material in Element.

        The order of thickness should correspond to the order of the Material objects in
        material_lst.
        """
        return self._layer_thickness_lst

    @layer_thickness_lst.setter
    def layer_thickness_lst(self, value):
        assert all(v > 0 for v in value), 'Every value in layer_thickness_lst '
        'must be greater than 0.'
        self._layer_thickness_lst = value

    @property
    def material_lst(self):
        """Get or set list of Material objects in the element.

        The order of Material objects should correspond to the order of the thickness in
        layer_thickness_lst.
        """
        return self._material_lst

    @material_lst.setter
    def material_lst(self, value):
        assert all(isinstance(v, Material) for v in value), 'Every item in '
        'in material_lst must be a Material object.'
        self._material_lst = value

    @property
    def vegcoverage(self):
        """Get or set fraction of vegetation coverage on Element."""
        return self._vegcoverage

    @vegcoverage.setter
    def vegcoverage(self, value):
        self._vegcoverage = float_in_range(value, 0, 1, 'vegcoverage')

    @property
    def t_init(self):
        """Get or set initial temperature of Element [K]."""
        return self._t_init

    @t_init.setter
    def t_init(self, value):
        self._t_init = float_in_range(value, mi=0, input_name='t_init')

    @property
    def horizontal(self):
        """Get or set boolean value indicating if Element is horizontal or not."""
        return self._horizontal

    @horizontal.setter
    def horizontal(self, value):
        self._horizontal = bool(value)

    @property
    def name(self):
        """Get or set text string for name of Element."""
        return self._name

    @classmethod
    def from_dict(cls, data):
        """Create a Element object from a dictionary.

        Args:
            data: A Element dictionary following the format below.

        .. code-block:: python

            {
            "albedo": 0.08,
            "emissivity": 0.92,
            "layer_thickness_lst": [0.0254, 0.0508, 0.0254],
            "material_lst": material_dict_lst,  # List of material dictionaries
            "vegcoverage": 0.5,
            "t_init": 293,
            "horizontal": False,
            "name": MassWall
            }
        """
        assert data['type'] == 'Element', 'Expected ' \
            'Element dictionary. Got {}.'.format(data['type'])

        materials = [Material.from_dict(m) for m in data['material_lst']]
        return cls(data['albedo'], data['emissivity'], data['layer_thickness_lst'],
                   materials, data['vegcoverage'], data['t_init'],
                   data['horizontal'], data['name'])

    def to_dict(self):
        """Element dictionary representation."""
        base = {'type': 'Element'}
        base['albedo'] = self.albedo
        base['emissivity'] = self.emissivity
        base['layer_thickness_lst'] = self.layer_thickness_lst
        base['material_lst'] = [m.to_dict() for m in self.material_lst]
        base['vegcoverage'] = self.vegcoverage
        base['t_init'] = self.t_init
        base['horizontal'] = self.horizontal
        base['name'] = self.name
        return base

    def SurfFlux(self, forc, parameter, simTime, humRef, tempRef, windRef, boundCond,
                 intFlux):
        """Calculate net heat flux, and update element layer temperatures."""

        # Calculated per unit area [m^2]

        # dens: air density (kgd m-3)
        dens = forc.pres / (1000 * 0.287042 * tempRef *
                            (1. + 1.607858 * humRef))
        # convection coef (ref: uwg, eq. 12))
        self.aeroCond = 5.8 + 3.7 * windRef

        if self.horizontal:
            # For roof, mass, road
            if not is_near_zero(self.waterStorage) and self.waterStorage > 0.0:
                # Evaporation [m s-1], Film water & soil latent heat
                # Note: in the current uwg code, latent heat from evapotranspiration,
                # stagnant water, or anthropogenic sources is not modelled due to the
                # difficulty of validation, and lack of reliability of precipitation
                # data from EPW files.Therefore this condition is never run because all
                # elements have had their waterStorage hardcoded to 0.
                qtsat = self.qsat([self.layerTemp[0]], [
                                  forc.pres], parameter)[0]
                eg = self.aeroCond * parameter.colburn * dens * \
                    (qtsat - humRef) / parameter.waterDens / parameter.cp
                self.waterStorage = \
                    min(self.waterStorage + simTime.dt * (forc.prec - eg),
                        parameter.wgmax)
                self.waterStorage = max(self.waterStorage, 0.)  # [m]
            else:
                eg = 0.
            soilLat = eg * parameter.waterDens * parameter.lv

            if simTime.month < parameter.vegStart and simTime.month > parameter.vegEnd:
                # Winter, no veg
                self.solAbs = (1.0 - self.albedo) * self.solRec  # (W m-2)
                vegLat = 0.
                vegSen = 0.
            else:
                # Summer, veg
                self.solAbs = (
                    ((1.0 - self.vegcoverage) * (1. - self.albedo) + self.vegcoverage *
                     (1.0 - parameter.vegAlbedo)) * self.solRec)
                try:
                    # if road compute grass/tree fractions seperately
                    vegLat = (
                        self.grasscoverage * (1.0 - parameter.vegAlbedo) *
                        parameter.grassFLat * self.solRec)
                    vegLat += (
                        self.treecoverage * (1.0 - parameter.vegAlbedo) *
                        parameter.treeFLat * self.solRec)
                    vegSen = (
                        self.grasscoverage * (1.0 - parameter.vegAlbedo) *
                        (1.0 - parameter.grassFLat) * self.solRec)
                    vegSen += (
                        self.treecoverage * (1.0 - parameter.vegAlbedo) *
                        (1.0 - parameter.treeFLat) * self.solRec)
                except AttributeError:
                    # for all other Elements use veg fraction w/ grassFLat
                    vegLat = (
                        self.vegcoverage * (1.0 - parameter.vegAlbedo) *
                        parameter.grassFLat * self.solRec)
                    vegSen = (
                        self.vegcoverage * (1.0 - parameter.vegAlbedo) *
                        (1.0 - parameter.grassFLat) * self.solRec)

            self.lat = soilLat + vegLat

            # Sensible & net heat flux
            self.sens = vegSen + self.aeroCond * (self.layerTemp[0] - tempRef)
            self.flux = self.solAbs + self.infra - \
                self.lat - self.sens  # [W m-2]

        else:
            # For vertical surfaces (wall)
            self.solAbs = (1.0 - self.albedo) * self.solRec
            self.lat = 0.0

            # Sensible & net heat flux
            self.sens = self.aeroCond * (self.layerTemp[0] - tempRef)
            self.flux = self.solAbs + self.infra - \
                self.lat - self.sens  # [W m-2]

        self.layerTemp = \
            self.Conduction(simTime.dt, self.flux, boundCond,
                            forc.deepTemp, intFlux)
        self.T_ext = self.layerTemp[0]
        self.T_int = self.layerTemp[-1]

    def Conduction(self, dt, flx1, bc, temp2, flx2):
        """Solve the conductance of heat based on of the element layers.

        Args:
            dt: Simulation time step in seconds.
            flx1: Net heat flux on surface [W m-2]
            bc: Boundary condition parameter (1 or 2)
            temp2: Deep soil temperature (ave of air temperature) [K]
            flx2: Surface flux (sum of absorbed, emitted, etc.) [W m-2]

        Returns:
            A 1d vector of element layer temperatures.
        """
        t = self.layerTemp          # vector of layer temperatures (K)
        # vector of layer volumetric heat (J m-3 K-1)
        hc = self.layerVolHeat
        # vector of layer thermal conductivities (W m-1 K-1)
        tc = self.layerThermalCond
        d = self.layer_thickness_lst     # vector of layer thicknesses (m)

        fimp = 0.5                  # implicit coefficient
        fexp = 0.5                  # explicit coefficient
        num = len(t)                # number of layers

        # Mean thermal conductivity over distance between 2 layers (W/mK)
        tcp = [0 for x in range(num)]
        # Thermal capacity times layer depth (J/m2K)
        hcp = [0 for x in range(num)]
        # lower, main, and upper diagonals
        za = [[0 for y in range(3)] for x in range(num)]
        # RHS
        zy = [0 for x in range(num)]

        # --------------------------------------------------------------------------
        # Define the column vectors for heat capactiy and conductivity
        hcp[0] = hc[0] * d[0]
        for j in range(1, num):
            tcp[j] = 2. / (d[j-1] / tc[j-1] + d[j] / tc[j])
            hcp[j] = hc[j] * d[j]

        # --------------------------------------------------------------------------
        # Define the first row of za matrix, and RHS column vector
        za[0][0] = 0.
        za[0][1] = hcp[0] / dt + fimp * tcp[1]
        za[0][2] = -fimp * tcp[1]
        zy[0] = hcp[0] / dt*t[0] - fexp * tcp[1] * (t[0] - t[1]) + flx1

        # --------------------------------------------------------------------------
        # Define other rows
        for j in range(1, num - 1):
            za[j][0] = fimp * (-tcp[j])
            za[j][1] = hcp[j] / dt + fimp * (tcp[j] + tcp[j+1])
            za[j][2] = fimp * (-tcp[j+1])
            zy[j] = hcp[j] / dt * t[j] + fexp * \
                (tcp[j] * t[j - 1] - tcp[j] * t[j] -
                 tcp[j+1] * t[j] + tcp[j+1] * t[j+1])

        # --------------------------------------------------------------------------
        # Boundary conditions
        if is_near_zero(bc - 1.0):
            # heat flux
            za[num-1][0] = fimp * (-tcp[num-1])
            za[num-1][1] = hcp[num-1] / dt + fimp * tcp[num-1]
            za[num-1][2] = 0.
            zy[num-1] = hcp[num-1] / dt * t[num-1] + fexp * tcp[num-1] * \
                (t[num-2] - t[num-1]) + flx2
        elif is_near_zero(bc - 2.0):
            # deep-temperature
            za[num-1][0] = 0.
            za[num-1][1] = 1.
            za[num-1][2] = 0.
            zy[num-1] = temp2
        else:
            raise Exception('Error during conduction calculation. Check input '
                            'parameters in the Conduction routine.')

        # --------------------------------------------------------------------------
        zx = Element.invert(num, za, zy)
        # t(:) = zx(:)
        return zx  # return zx as 1d vector of templayers

    def qsat(self, temp, pres, parameter):
        """Calculate vector of saturation humidity from air pressure and layer temperatures.

        Args:
            temp: List of layer temperatures [K].
            pres: Pressure (at current timestep) [Pa].
            parameter: Parameter object with geographic parameters.

        Returns:
            List of saturation humidity values.
        """
        gamw = (parameter.cl - parameter.cpv) / parameter.rv
        betaw = (parameter.lvtt / parameter.rv) + (gamw * parameter.tt)
        alpw = math.log(parameter.estt) + (betaw / parameter.tt) + \
            (gamw * math.log(parameter.tt))
        work2 = parameter.r / parameter.rv
        foes_lst = [0 for i in range(len(temp))]
        work1_lst = [0 for i in range(len(temp))]
        qsat_lst = [0 for i in range(len(temp))]

        for i in range(len(temp)):
            # saturation vapor pressure
            foes_lst[i] = math.exp(
                alpw - (betaw / temp[i]) - (gamw * math.log(temp[i])))
            work1_lst[i] = foes_lst[i] / pres[i]
            # saturation humidity
            qsat_lst[i] = work2 * work1_lst[i] / \
                (1. + ((work2 - 1.) * work1_lst[i]))

        return qsat_lst

    @staticmethod
    def invert(nz, A, C):
        """Invert and solve tridiagonal matrix.

        Given A * X = C, solves for X, where
            a(*, 1) lower diagonal (Ai, i-1)
            a(*, 2) principal diagonal (Ai, i)
            a(*, 3) upper diagonal (Ai, i+1)

        Args:
            nz: number of layers.
            A: A matrix to invert.
            C: Result from A * X

        Returns:
            Solved matrix X from matrix inversion.
        """

        X = [0 for i in range(nz)]

        for i in reversed(range(nz-1)):
            C[i] = C[i] - A[i][2] * C[i+1] / A[i+1][1]
            A[i][1] = A[i][1] - A[i][2] * A[i+1][0] / A[i+1][1]

        for i in range(1, nz, 1):
            C[i] = C[i] - A[i][0] * C[i-1] / A[i-1][1]

        for i in range(nz):
            X[i] = C[i] / A[i][1]

        return X

    def __repr__(self):
        rval = round(sum([1.0 / t for t in self.layerThermalCond]), 2)
        return 'Element,\n name: {}\n emissivity: {}\n albedo: {}\n R value: ' \
            '{}\n vegcoverage: {}\n horizontal: {}\n layer_thickness_lst: {}\n ' \
            'layerTemp: {}'.format(
                self.name, self.emissivity, self.albedo, rval, self.vegcoverage,
                bool(self.horizontal), self.layer_thickness_lst, self.layerTemp)



"""Collection of useful methods."""
from csv import reader as csv_reader
import sys

try:
    range = xrange
except NameError:
    pass


try:
    import math
    INFPOS = math.inf
    INFNEG = -1 * math.inf
except AttributeError:
    # python 2
    INFPOS = float('inf')
    INFNEG = float('-inf')


# DOE References
REF_BLDTYPE = ('fullservicerestaurant', 'hospital', 'largehotel', 'largeoffice',
               'medoffice', 'midriseapartment', 'outpatient', 'primaryschool',
               'quickservicerestaurant', 'secondaryschool', 'smallhotel',
               'smalloffice', 'standaloneretail', 'stripmall', 'supermarket',
               'warehouse')
REF_BUILTERA = ('pre80', 'pst80', 'new')
REF_ZONETYPE = ('1A', '2A', '2B', '3A', '3B-CA', '3B', '3C', '4A', '4B', '4C', '5A',
                '5B', '6A', '6B', '7', '8')
REF_BLDTYPE_SET = {'fullservicerestaurant', 'hospital', 'largehotel', 'largeoffice',
                   'medoffice', 'midriseapartment', 'outpatient', 'primaryschool',
                   'quickservicerestaurant', 'secondaryschool', 'smallhotel',
                   'smalloffice', 'standaloneretail', 'stripmall', 'supermarket',
                   'warehouse'}
REF_ZONETYPE_SET = {'1A', '2A', '2B', '3A', '3B-CA', '3B', '3C', '4A', '4B', '4C', '5A',
                    '5B', '6A', '6B', '7', '8'}
REF_BUILTERA_SET = {'pre80', 'pst80', 'new'}


def is_near_zero(num, eps=1e-10):
    return abs(float(num)) < eps


def read_csv(file_path):
    """Open csv file and read.

    Args:
        file_path: Text string for file path.

    Returns:
        List of file lines as str type.
    """
    if sys.version_info[0] >= 3:
        file_ = open(file_path, "r", errors='ignore')
    else:
        file_ = open(file_path, "r")

    gen_ = csv_reader(file_, delimiter=",")
    L = [r for r in gen_]
    file_.close()
    return L


def str2fl(x):
    """Recurses through lists and converts lists of string to float

    Args:
        x: string or list of strings
    """
    def helper_to_fl(s_):
        """Deals with odd string imports converts to float"""
        if s_ == "":
            return "null"
        elif "," in s_:
            s_ = s_.replace(",", "")

        try:
            return float(s_)
        except (ValueError, TypeError):
            return (s_)

    fl_lst = []
    if isinstance(x[0], str):
        # Check if list of strings then conversion
        for xi in range(len(x)):
            fl_lst.append(helper_to_fl(x[xi]))
    elif isinstance(x[0], list):
        # Check if list of lists, then recurse
        for xi in range(len(x)):
            fl_lst.append(str2fl(x[xi]))
    else:
        return False

    return fl_lst


def float_in_range(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be between minimum and maximum."""
    assert mi <= value <= ma, 'Input number {} must be between {} and {}. ' \
        'Got {}'.format(input_name, mi, ma, value)
    return value


def float_in_range_excl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than minimum and less than maximum."""
    assert mi < value < ma, 'Input number {} must be greater than {} ' \
        'and less than {}. Got {}'.format(input_name, mi, ma, value)
    return value


def float_in_range_excl_incl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than minimum and less than/equal to maximum."""
    assert mi < value <= ma, 'Input number {} must be greater than {} and less than ' \
        'or equal to {}. Got {}'.format(input_name, mi, ma, value)
    return value


def float_in_range_incl_excl(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check a float value to be greater than/equal to minimum and less than maximum."""
    assert mi <= value < ma, 'Input number {} must be greater than or equal to {} ' \
        'and less than {}. Got {}'.format(input_name, mi, ma, value)
    return value


def int_in_range(value, mi=INFNEG, ma=INFPOS, input_name=''):
    """Check an integer value to be between minimum and maximum."""
    number = int(value)
    assert mi <= number <= ma, 'Input integer {} must be between {} and {}. ' \
        'Got {}.'.format(input_name, mi, ma, value)
    return number


def float_positive(value, input_name=''):
    """Check a float value to be positive."""
    return float_in_range(value, 0, INFPOS, input_name)


def int_positive(value, input_name=''):
    """Check if an integer value is positive."""
    return int_in_range(value, 0, INFPOS, input_name)



import pytest
from uwg import psychrometrics


def test_psychrometric_float_point():

    # Input values
    Tdb_in = 297.5311337413935
    w_in = 0.018576773131376
    P = 10090

    Tdb, w, phi, h, Tdp, v = psychrometrics.psychrometrics(Tdb_in, w_in, P)

    assert Tdb == pytest.approx(24.381133741393512, abs=1e-15)   # Tdb [C]
    assert w == pytest.approx(0.018576773131376, abs=1e-15)      # W [kgv/kgd]
    assert phi == pytest.approx(9.581555922752541, abs=1e-15)    # RH Pw/Pws*100
    assert h == pytest.approx(7.183036653518451e+04, abs=1e-15)  # Enthalpy [J/kgd]
    assert Tdp == pytest.approx(-10.012150181172135, abs=1e-15)  # Wet bulb temp [C]
    assert v == pytest.approx(8.717031296493113, abs=1e-15)      # Spec. vol [m3/kga]


def test_psychrometric_simple_1():
    # Really simple, coarse tests
    # Input values
    Tdb_in = 20.0 + 273.15
    w_in = 0.002
    P = 101325.0

    Tdb, w, phi, h, Tdp, v = psychrometrics.psychrometrics(Tdb_in, w_in, P)

    # Tests for 20, 0.002, atmosphere
    assert Tdb + 273.15 == pytest.approx(Tdb_in, True, 1e-14)    # Tdb [C]
    assert w == pytest.approx(w_in, 1e-14)                       # W [kgv/kga]
    assert phi == pytest.approx(13., 1)                          # RH [%}
    assert h / 1000. == pytest.approx(25., 1)                    # Enthalpy [J/kga]
    assert v == pytest.approx(0.83, 1e-2)                        # Spec. vol [m^3 kg-1]


def test_psychrometric_simple_2():
    # Really simple, coarse tests
    # Input values
    Tdb_in = 40.0 + 273.15
    w_in = 0.009
    P = 101325.0

    Tdb, w, phi, h, Tdp, v = psychrometrics.psychrometrics(Tdb_in, w_in, P)

    # Tests for 40, 0.009, atmosphere
    assert Tdb+273.15 == pytest.approx(Tdb_in, True, 1e-14)      # Tdb [C]
    assert w == pytest.approx(w_in, 1e-14)                       # W [kgv/kga]
    assert phi == pytest.approx(19.5, 1e-1)                      # RH [%}
    assert h / 1000. == pytest.approx(63., 1)                    # Enthalpy [J/kga]
    assert v == pytest.approx(0.9, 1e-1)                         # Spec. vol [m^3 kg-1]




import os
from uwg import UWG
import math
import logging

TEST_DIR = os.path.dirname(__file__)
DEFAULT_EPW_PATH = os.path.join(
    TEST_DIR, 'epw', 'SGP_Singapore.486980_IWEC.epw')
DEFAULT_PARAM_PATH = os.path.join(
    TEST_DIR, 'parameters', 'initialize_singapore.uwg')
NEW_DIR = os.path.join(TEST_DIR, 'epw_uwg')
MATLAB_DIR = os.path.join(TEST_DIR, 'matlab_ref')


def calculate_tolerance(x, p):
    if abs(float(x)) > 1e-15:
        tol = 1 * 10 ** -(p - 1 - int(math.log10(abs(x))))
    else:
        tol = 1e-15
    return tol


def auto_setup_uwg(epw_path=DEFAULT_EPW_PATH, param_path=DEFAULT_PARAM_PATH,
                   log_file_name=None, log_level=None, uwg_param_dir=None):
    """Set up uwg object from initialize.uwg."""
    if log_file_name and log_level:
        setup_log_file(log_file_name, log_level)

    if param_path:
        model = UWG.from_param_file(
            param_path, epw_path=epw_path, new_epw_dir=NEW_DIR)
    else:
        model = UWG(epw_path, new_epw_dir=NEW_DIR)

    # Increase precision for testing
    model.epw_precision = 16

    return model


def set_input_manually(model):
    """Assign everything manually from ../resources/initialize_singapore.uwg"""

    # Define Simulation and Weather parameters
    model.month = 1.
    model.day = 1.
    model.nday = 31.
    model.dtsim = 300.
    model.dtweather = 3600.

    # HVAC system and internal laod
    model.autosize = 0.
    model.sensocc = 100.
    model.latfocc = 0.3
    model.radfocc = 0.2
    model.radfequip = 0.5
    model.radflight = 0.7

    # Define Urban microclimate parameters
    model.h_ubl1 = 1000.
    model.h_ubl2 = 80.
    model.h_ref = 150.
    model.h_temp = 2.
    model.h_wind = 10.
    model.c_circ = 1.2
    model.c_exch = 1.
    model.maxday = 150.
    model.maxnight = 20.
    model.windmin = 1.
    model.h_obs = 0.1

    # Urban characteristics
    model.bldheight = 10.
    model.h_mix = 1.
    model.blddensity = 0.5
    model.vertohor = 0.8
    model.charlength = 1000.
    model.albroad = 0.1
    model.droad = 0.5
    model.sensanth = 20.

    # Define optional Building characteristics
    model.bld = [('largeoffice', 'pst80', 0.4),
                 ('midriseapartment', 'pst80', 0.6)]

    # climate Zone
    model.zone = '1A'

    # Vegetation parameters
    model.vegstart = 4.0
    model.vegend = 10.0
    model.grasscover = 0.1
    model.treecover = 0.1
    model.albveg = 0.25
    model.rurvegcover = 0.9
    model.latgrss = 0.4
    model.lattree = 0.6

    # Define Traffic schedule
    model.schtraffic = [
        [0.2, 0.2, 0.2, 0.2, 0.2, 0.4, 0.7, 0.9, 0.9, 0.6, 0.6, 0.6, 0.6, 0.6, 0.7, 0.8,
         0.9, 0.9, 0.8, 0.8, 0.7, 0.3, 0.2, 0.2],  # Weekday
        [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.6, 0.7,
         0.7, 0.7, 0.7, 0.5, 0.4, 0.3, 0.2, 0.2],  # Saturday
        [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4,
         0.4, 0.4, 0.4, 0.4, 0.3, 0.3, 0.2, 0.2]]  # Sunday

    # Define Road (Assume 0.5m of asphalt)
    model.kroad = 1.
    model.croad = 1600000.

    # Optional parameters are not defined as they are already
    # initialized as None
    return model


def setup_open_matlab_ref(matlab_class_dir, matlab_ref_file_path):
    """ Open the matlab reference file """
    def convert_type(x):
        try:
            return float(x)
        except ValueError:
            return x

    matlab_path = os.path.join(
        MATLAB_DIR, matlab_class_dir, matlab_ref_file_path)
    if not os.path.exists(matlab_path):
        raise Exception("Failed to open {}!".format(matlab_path))

    # Open matlab file, read lines and return values as python simple datatype
    matlab_file = open(matlab_path, 'r')
    uwg_matlab_val_ = [convert_type(x) for x in matlab_file.readlines()]
    matlab_file.close()
    return uwg_matlab_val_


def setup_log_file(log_file_name, log_level=None):
    """
    https://fangpenlin.com/posts/2012/08/26/good-logging-practice-in-python/

    # LOGGER LEVELS
    DEBUG                               # Detailed information
    INFO                                # Confirm things are working as expected
    WARNING                             # Indication something unexpected has happened
    ERROR                               # Error occurs
    CRITICAL                            # Program may not be able to run
    """

    log_file_path = os.path.join(TEST_DIR, "..", "logs", log_file_name)

    if log_level is None:
        log_level = logging.DEBUG  # Default is set to debug for everything

    logging.basicConfig(level=log_level,
                        filename=log_file_path,
                        filemode="w")

    # To stream to console
    # logging.basicConfig(level=logging.DEBUG)



"""Test for schdef.py"""

import pytest
from uwg import SchDef


DEFAULT_WEEK = [
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.4, 0.7, 0.9, 0.9, 0.6, 0.6, 0.6, 0.6, 0.6,
     0.7, 0.8, 0.9, 0.9, 0.8, 0.8, 0.7, 0.3, 0.2, 0.2],  # Weekday
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5,
        0.6, 0.7, 0.7, 0.7, 0.7, 0.5, 0.4, 0.3, 0.2, 0.2],  # Saturday
    [0.2, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4,
        0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.3, 0.3, 0.2, 0.2]]  # Sunday


def test_schdef_init():
    """Test SchDef init method."""

    with pytest.raises(AssertionError):
        SchDef(elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK, occ=DEFAULT_WEEK,
               cool=DEFAULT_WEEK, heat=DEFAULT_WEEK, swh={}, q_elec=18.9, q_gas=3.2,
               q_light=18.9, n_occ=0.12, vent=0.0013, v_swh=0.2846,
               bldtype='fullservicerestaurant', builtera='pre80')

    with pytest.raises(AssertionError):
        testweek = [[0.1] * 24] * 2
        SchDef(elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK, occ=DEFAULT_WEEK,
               cool=DEFAULT_WEEK, heat=testweek, swh=DEFAULT_WEEK, q_elec=18.9,
               q_gas=3.2, q_light=18.9, n_occ=0.12, vent=0.0013, v_swh=0.2846,
               bldtype='fullservicerestaurant', builtera='pre80')

    with pytest.raises(AssertionError):
        testweek = [[0.1] * 24] * 2
        testweek += [0.1] * 23
        SchDef(elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK, occ=DEFAULT_WEEK,
               cool=DEFAULT_WEEK, heat=testweek, swh=DEFAULT_WEEK, q_elec=18.9,
               q_gas=3.2, q_light=18.9, n_occ=0.12, vent=0.0013, v_swh=0.2846,
               bldtype='fullservicerestaurant', builtera='pre80')

    with pytest.raises(AssertionError):
        testweek = [[0.1] * 24] * 3
        testweek[2][23] = 'a'
        SchDef(elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK, occ=DEFAULT_WEEK,
               cool=DEFAULT_WEEK, heat=testweek, swh=DEFAULT_WEEK, q_elec=18.9,
               q_gas=3.2, q_light=18.9, n_occ=0.12, vent=0.0013, v_swh=0.2846,
               bldtype='fullservicerestaurant', builtera='pre80')

    # init
    sch = SchDef(elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK,
                 occ=DEFAULT_WEEK, cool=DEFAULT_WEEK, heat=DEFAULT_WEEK,
                 swh=DEFAULT_WEEK, q_elec=18.9, q_gas=3.2, q_light=18.9, n_occ=0.12,
                 vent=0.0013, v_swh=0.2846, bldtype='fullservicerestaurant',
                 builtera='pre80')

    # test __repr__
    sch.__repr__()


def test_schdef_dict():
    """Test SchDef dict methods."""

    # init
    testsch = SchDef(
        elec=DEFAULT_WEEK, gas=DEFAULT_WEEK, light=DEFAULT_WEEK, occ=DEFAULT_WEEK,
        cool=DEFAULT_WEEK, heat=DEFAULT_WEEK, swh=DEFAULT_WEEK, q_elec=18.9,
        q_gas=3.2, q_light=18.9, n_occ=0.12, vent=0.0013, v_swh=0.2846,
        bldtype='fullservicerestaurant', builtera='pre80')

    # make dict
    schdict = testsch.to_dict()

    # test if dict and from_dict
    assert isinstance(schdict, dict)

    testsch2 = SchDef.from_dict(schdict)

    assert testsch.elec == testsch2.elec
    assert testsch.gas == testsch2.gas
    assert testsch.light == testsch2.light
    assert testsch.occ == testsch2.occ
    assert testsch.cool == testsch2.cool
    assert testsch.heat == testsch2.heat
    assert testsch.swh == testsch2.swh

    with pytest.raises(AssertionError):
        schdict['type'] = 'BemDef'
        SchDef.from_dict(schdict)



"""Test UCMDef object."""

try:
    range = xrange
except NameError:
    pass

import pytest
import math
from .test_base import auto_setup_uwg, setup_open_matlab_ref, calculate_tolerance


def test_ucm_init():
    """Test ucm constructor."""

    testuwg = auto_setup_uwg()
    testuwg.generate()

    # Because we are not using latAnthrop, this value can be None in uwg.
    # So we will hardcode value for now.
    testuwg.UCM.latAnthrop = 2.0

    # Get uwg_python values
    uwg_python_val = [
        testuwg.UCM.h_mix,         # waste heat mix into canyon ratio
        testuwg.UCM.bldHeight,     # average building height (m)
        testuwg.UCM.verToHor,      # vertical-to-horizontal urban area ratio
        testuwg.UCM.bldDensity,    # horizontal building density (footprint)
        testuwg.UCM.treeCoverage,  # horizontal tree coverage (footprint)
        testuwg.UCM.sensAnthrop,   # sensible heat anthropogenic (W m-2)
        testuwg.UCM.latAnthrop,    # latent heat anthropogenic (W m-2)
        testuwg.UCM.roadShad,      # shadow fraction of road
        testuwg.UCM.bldWidth,      # building width (m)
        testuwg.UCM.canWidth,      # canyon width (m)
        testuwg.UCM.canAspect,     # canyon aspect ratio
        testuwg.UCM.roadConf,      # road sky view factor
        testuwg.UCM.wallConf,      # wall sky view factor
        testuwg.UCM.facArea,       # facade area (m2)
        testuwg.UCM.roadArea,      # road area (m2)
        testuwg.UCM.roofArea,      # roof area (m2)
        testuwg.UCM.canTemp,       # canyon air temperature (db) (K)
        testuwg.UCM.roadTemp,      # avg road temperature
        testuwg.UCM.canHum,        # canyon specific humidity (kgv kgda-1)
        testuwg.UCM.ublWind,       # urban boundary layer wind velocity (m s-1)
        testuwg.UCM.canWind,       # urban canyon wind velocity (m s-1)
        testuwg.UCM.ustar,         # friction velocity (m s-1)
        testuwg.UCM.ustarMod,      # modified friction velocity (m s-1)
        testuwg.UCM.z0u,           # urban roughness length (m)
        testuwg.UCM.l_disp,        # urban displacement length (m)
        testuwg.UCM.alb_wall,      # albedo of wall
        testuwg.UCM.facAbsor,      # absorptivity of wall
        testuwg.UCM.roadAbsor,     # average road absorptivity
        testuwg.UCM.sensHeat]      # urban sensible heat [W m-2]

    uwg_matlab_val = testuwg = setup_open_matlab_ref(
        'matlab_ucm', 'matlab_ref_ucm_init.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)
    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(uwg_python_val[i], 15.0)
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], tol), \
            'error at index={}'.format(i)


def test_ucm_ucmodel():
    """Test ucm ucmodel."""

    testuwg = auto_setup_uwg()

    # Test Jan 1 (winter, no vegetation coverage)
    testuwg.month = 1
    testuwg.day = 1
    testuwg.nday = 1

    # set_input
    testuwg._read_epw()
    testuwg._compute_BEM()
    testuwg._compute_input()

    # In order to avoid integration effects. Test only first time step
    # Subtract timestep to stop at 300 sec
    testuwg.simTime.nt -= (23 * 12 + 11)

    # Run simulation
    testuwg._hvac_autosize()
    testuwg.simulate()

    # check date
    assert testuwg.simTime.month == 1
    assert testuwg.simTime.day == 1
    assert testuwg.simTime.secDay == pytest.approx(300.0, abs=1e-15)

    # Get uwg values
    uwg_python_val = [
        # heat load building
        testuwg.UCM.Q_wall,        # convective sensible heat flux from building roof
        testuwg.UCM.Q_window,      # sensible heat flux from window (via U-factor)
        testuwg.UCM.Q_hvac,        # sensible heat flux from HVAC waste
        testuwg.UCM.ElecTotal,     # total electricity consumption of urban area
        testuwg.UCM.GasTotal,      # total gas consumption of urban area
        # heat load road/canyon/ubl
        testuwg.UCM.Q_road,        # convective sensible heat flux from road
        testuwg.UCM.Q_traffic,     # net sensible heat flux from traffic
        testuwg.UCM.Q_vent,        # convective heat exchange from vent/infil
        testuwg.UCM.Q_ubl,         # convective heat exchange with UBL layer
        # Building temperature
        testuwg.UCM.roofTemp,      # average roof temp (K)
        testuwg.UCM.wallTemp,      # average wall temp (K)
        # Sensible heat
        testuwg.UCM.treeSensHeat,  # sensible heat from trees
        testuwg.UCM.sensHeat,      # urban sensible heat
        testuwg.UCM.canTemp        # canyon air temp (K)
    ]

    # Get uwg_matlab values
    uwg_matlab_val = setup_open_matlab_ref('matlab_ucm', 'matlab_ref_ucm_ucmodel.txt')

    # matlab ref checking
    assert len(uwg_matlab_val) == len(uwg_python_val)

    for i in range(len(uwg_matlab_val)):
        tol = calculate_tolerance(abs(uwg_python_val[i]), 15.0)
        tol = 10 ** (math.log10(tol) + 2)  # Lowering tolerance by two orders of magnitude
        assert uwg_python_val[i] == pytest.approx(uwg_matlab_val[i], abs=tol), \
            'error at index={}'.format(i)




"""Test the readDOE module."""

import os
import math
import copy
from functools import reduce
import pytest

from .test_base import calculate_tolerance, MATLAB_DIR
from uwg.readDOE import readDOE


MATLAB_DIR = os.path.join(MATLAB_DIR, 'matlab_readDOE')

# Test matrix of ref data as nested nested lists [16, 3, 16]:

# refDOE = Building objs
# Schedule = SchDef objs
# refBEM = BEMDef

# where:
#     [16,3,16] is Type = 1-16, Era = 1-3, climate zone = 1-16
#     i.e.
#     Type: FullServiceRestaurant, Era: Pre80, Zone: 6A Minneapolis
# Nested tree:
# [TYPE_1:
#     ERA_1:
#         CLIMATE_ZONE_1
#         ...
#         CLIMATE_ZONE_16
#     ERA_2:
#         CLIMATE_ZONE_1
#         ...
#         CLIMATE_ZONE_16
#     ...
#     ERA_3:
#         CLIMATE_ZONE_1
#         ...
#         CLIMATE_ZONE_16]


def test_refDOE():
    """Tests for refDOE (Building class)"""

    # Run readDOE.py
    refBEM, Schedule = readDOE(serialize_output=False)

    bldlst = ['floorHeight', 'intHeat', 'intHeatNight', 'intHeatDay', 'intHeatFRad',
              'intHeatFLat', 'infil', 'vent', 'glazingRatio', 'uValue', 'shgc',
              'condType', 'cop', 'coolSetpointDay', 'coolSetpointNight',
              'heatSetpointDay', 'heatSetpointNight', 'coolCap', 'heatEff', 'mSys',
              'indoorTemp', 'indoorHum', 'heatCap', 'copAdj', 'fanMax']

    for bi in range(len(bldlst)):
        bldid = bldlst[bi]
        matlab_path = os.path.join(MATLAB_DIR, 'matlab_ref_{}_.txt'.format(bldid))

        # check file exists
        if not os.path.exists(matlab_path):
            raise Exception('Failed to open {}!'.format(matlab_path))

        matlab_file = open(matlab_path, 'r')

        refDOE = refBEM

        assert len(refDOE) == pytest.approx(16.0, abs=1e-3)

        for bldType in range(16):         # bldType
            assert len(refDOE[bldType]) == pytest.approx(3.0, abs=1e-3)

            for bldEra in range(3):        # bltEra
                assert len(refDOE[bldType][bldEra]) == pytest.approx(16.0, abs=1e-3)

                for climateZone in range(16):  # ZoneType

                    matlab_ref_value = matlab_file.readline()

                    if bldid != 'condType':
                        matlab_ref_value = float(matlab_ref_value)
                        tol = calculate_tolerance(matlab_ref_value, 14.0)
                    else:
                        matlab_ref_value = 'AIR'

                    bldg = refDOE[bldType][bldEra][climateZone].building

                    # run tests
                    if bldid == 'floorHeight':
                        assert bldg.floor_height == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'intHeat':
                        assert bldg.int_heat == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'intHeatNight':
                        assert bldg.int_heat_night == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'intHeatDay':
                        assert bldg.int_heat_day == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'intHeatFRad':
                        assert bldg.int_heat_frad == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'intHeatFLat':
                        assert bldg.int_heat_flat == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'infil':
                        assert bldg.infil == pytest.approx(matlab_ref_value,  abs=tol),  \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'vent':
                        assert bldg.vent == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'glazingRatio':
                        assert bldg.glazing_ratio == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'uValue':
                        assert bldg.u_value == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'shgc':
                        assert bldg.shgc == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'condType':
                        assert bldg.condtype == matlab_ref_value,  \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'cop':
                        assert bldg.cop == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'coolSetpointDay':
                        assert bldg.cool_setpoint_day == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'coolSetpointNight':
                        assert bldg.cool_setpoint_night == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'heatSetpointDay':
                        assert bldg.heat_setpoint_day == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'heatSetpointNight':
                        assert bldg.heat_setpoint_night == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'coolCap': # knocking the testing tolerance one order of magnitude down due to excel extraction problem
                        assert bldg.coolcap == pytest.approx(matlab_ref_value, abs=10 ** (math.log10(tol) + 1)), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'heatEff':
                        assert bldg.heateff == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'mSys':
                        assert bldg.msys == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'indoorTemp':
                        assert bldg.indoor_temp == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'indoorHum':
                        assert bldg.indoor_hum == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'heatCap': # knocking the testing tolerance one order of magnitude down due to excel extraction problem
                        assert bldg.heat_cap == pytest.approx(matlab_ref_value, abs=10 ** (math.log10(tol) + 1)), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'copAdj':
                        assert bldg.cop_adj == pytest.approx(matlab_ref_value, abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'canyon_fraction':
                        assert bldg.canyon_fraction == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif bldid == 'fanMax':
                        assert refBEM[bldType][bldEra][climateZone].building.FanMax == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

        matlab_file.close()


def test_refBEM():
    """Tests for refBEM (BEMDef class)."""

    refBEM, Schedule = readDOE(serialize_output=False)

    elementlst = [
        'albedo', 'emissivity', 'layerThickness', 'layerThermalCond', 'layerVolHeat',
        'vegCoverage', 'layerTemp', 'horizontal']

    bemlst = [
        ('mass', copy.deepcopy(elementlst)),
        ('wall', copy.deepcopy(elementlst)),
        ('roof', copy.deepcopy(elementlst))]

    for bemi in range(len(bemlst)):
        for ei in range(len(elementlst)):
            bemid = bemlst[bemi][0] + '_' + bemlst[bemi][1][ei]

            matlab_path = \
                os.path.join(MATLAB_DIR, 'matlab_ref_bemdef_{}.txt'.format(bemid))

            # check file exists
            if not os.path.exists(matlab_path):
                raise Exception('Failed to open {}!'.format(matlab_path))

            matlab_file = open(matlab_path, 'r')

            assert len(refBEM) == pytest.approx(16.0, abs=1e-3)

            for bldType in range(16):         # bldType
                assert len(refBEM[bldType]) == pytest.approx(3.0, abs=1e-3)

                for bldEra in range(3):        # bltEra
                    assert len(refBEM[bldType][bldEra]) == pytest.approx(16.0, abs=1e-3)

                    for climateZone in range(16):  # ZoneType
                        # next line
                        matlab_ref_value = float(matlab_file.readline())
                        tol = calculate_tolerance(matlab_ref_value, 15.0)

                        # run tests
                        if bemid == 'mass_albedo':
                            assert refBEM[bldType][bldEra][climateZone].mass.albedo == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_emissivity':
                            assert refBEM[bldType][bldEra][climateZone].mass.emissivity == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_layerThickness':
                            assert refBEM[bldType][bldEra][climateZone].mass.layer_thickness_lst[0] == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_layerThermalCond':
                            assert refBEM[bldType][bldEra][climateZone].mass.layerThermalCond[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_layerVolHeat':
                            assert refBEM[bldType][bldEra][climateZone].mass.layerVolHeat[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_vegCoverage':
                            assert refBEM[bldType][bldEra][climateZone].mass.vegcoverage == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_layerTemp':
                            assert refBEM[bldType][bldEra][climateZone].mass.layerTemp[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'mass_horizontal':
                            assert refBEM[bldType][bldEra][climateZone].mass.horizontal == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_albedo':
                            assert refBEM[bldType][bldEra][climateZone].wall.albedo == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_emissivity':
                            assert refBEM[bldType][bldEra][climateZone].wall.emissivity == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_layerThickness':
                            assert refBEM[bldType][bldEra][climateZone].wall.layer_thickness_lst[0] == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_layerThermalCond':
                            assert refBEM[bldType][bldEra][climateZone].wall.layerThermalCond[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_layerVolHeat':
                            assert refBEM[bldType][bldEra][climateZone].wall.layerVolHeat[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_vegCoverage':
                            assert refBEM[bldType][bldEra][climateZone].wall.vegcoverage == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_layerTemp':
                            assert refBEM[bldType][bldEra][climateZone].wall.layerTemp[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'wall_horizontal':
                            assert refBEM[bldType][bldEra][climateZone].wall.horizontal == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_albedo':
                            assert refBEM[bldType][bldEra][climateZone].roof.albedo == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_emissivity':
                            assert refBEM[bldType][bldEra][climateZone].roof.emissivity == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_layerThickness':
                            assert refBEM[bldType][bldEra][climateZone].roof.layer_thickness_lst[0] == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_layerThermalCond':
                            assert refBEM[bldType][bldEra][climateZone].roof.layerThermalCond[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_layerVolHeat':
                            assert refBEM[bldType][bldEra][climateZone].roof.layerVolHeat[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_vegCoverage':
                            assert refBEM[bldType][bldEra][climateZone].roof.vegcoverage == pytest.approx(matlab_ref_value, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_layerTemp':
                            assert refBEM[bldType][bldEra][climateZone].roof.layerTemp[0] == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        elif bemid == 'roof_horizontal':
                            assert refBEM[bldType][bldEra][climateZone].roof.horizontal == pytest.approx(1, abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                        else:
                            assert refBEM[bldType][bldEra][climateZone].building.FanMax == pytest.approx(matlab_ref_value,  abs=tol), \
                                'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

    matlab_file.close()


def test_Schedule():
    """Tests for Schedule (Schdef class)."""

    refBEM, Schedule = readDOE(serialize_output=False)

    schlst = ['Elec', 'Light', 'Gas', 'Occ', 'Cool', 'Heat', 'SWH', 'Qelec', 'Qlight',
              'Nocc', 'Qgas', 'Vent', 'Vswh']

    for si in range(len(schlst)):
        schid = schlst[si]

        # Define file path
        matlab_path = os.path.join(MATLAB_DIR,'matlab_ref_sch_{}_.txt'.format(schid))
        # check file exists
        if not os.path.exists(matlab_path):
            raise Exception('Failed to open {}!'.format(matlab_path))

        matlab_file = open(matlab_path,'r')

        assert len(Schedule) == pytest.approx(16.0, abs=1e-3)

        for bldType in range(1): # bldType
            assert len(Schedule[bldType]) == pytest.approx(3.0, abs=1e-3)

            for bldEra in range(1): # bltEra
                assert len(Schedule[bldType][bldEra]) == pytest.approx(16.0, abs=1e-3)

                for climateZone in range(1): # ZoneType

                    if schid in schlst[:7]: #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        matlab_ref_str = matlab_file.readline()
                        # replace [,] -> '' && split at ' '
                        matlab_ref_str = matlab_ref_str.replace('[','').replace(']','').replace(';','_').replace(' ', '_')
                        matlab_ref_str = ''.join(matlab_ref_str.split())
                        matlab_ref_str = matlab_ref_str.split('_')
                        matlab_ref_value = [float(x) for x in matlab_ref_str]
                        # can compare equality of flat lists [0.12, 4, 5,1.12, 5, 6] == pytest.approx([0.124, 4., 5, 1.125, 5, 6], abs=1e-2)
                        assert len(matlab_ref_value) == pytest.approx(24.0*3.,abs=1e-3)

                        # Calculate tolerances and set minimum one for all
                        tol = min([calculate_tolerance(x,15.0) for x in matlab_ref_value])
                    else:
                        matlab_ref_value = float(matlab_file.readline())
                        tol = calculate_tolerance(matlab_ref_value, 15.0)

                    if schid == 'Elec': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].elec)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Light': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].light)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Gas': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].gas)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Occ': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].occ)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Cool': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].cool)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Heat': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].heat)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'SWH': #3x24 matrix of schedule for SWH (WD,Sat,Sun)
                        flatten_sch = reduce(lambda x,y: x+y, Schedule[bldType][bldEra][climateZone].swh)
                        assert flatten_sch == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Qelec':
                        assert Schedule[bldType][bldEra][climateZone].q_elec == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Qlight':
                        assert Schedule[bldType][bldEra][climateZone].q_light == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Nocc':
                        assert Schedule[bldType][bldEra][climateZone].n_occ == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Qgas':
                        assert Schedule[bldType][bldEra][climateZone].q_gas == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Vent':
                        assert Schedule[bldType][bldEra][climateZone].vent == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

                    elif schid == 'Vswh':
                        assert Schedule[bldType][bldEra][climateZone].v_swh == pytest.approx(matlab_ref_value,  abs=tol), \
                            'btype={},era={},czone={}'.format(bldType+1, bldEra+1, climateZone+1)

    matlab_file.close()

